from abc import abstractmethod
import dash_bootstrap_components as dbc

from dash import Dash

from ...domain import Controller
from ...service import init_logger


class MetaTab:
    def __init__(self, app: Dash, metabo_controller: Controller):
        self._logger = init_logger()
        self.metabo_controller = metabo_controller
        self.app = app
        self._registerCallbacks()

    @abstractmethod
    def getLayout(self) -> dbc.Tab:
        pass

    @abstractmethod
    def _registerCallbacks(self) -> None:
        pass



from typing import List, Tuple, Union

import pandas as pd
from sklearn.model_selection import train_test_split

from . import MetaData
from ..service import Utils, init_logger
import numpy as np


class SplitGroup:
    def __init__(self, metadata: MetaData, selected_targets: List[str], train_test_proportion: float,
                 number_of_splits: int, classes_design: dict, pairing_column: str, 
                 uniq_sample_id: List[str], balance_correction: int = 0,
                 classes_repartition: Union[dict, None] = None,
                 test_split_seed: Union[int,None] = None):
        self._logger = init_logger()
        self._metadata = metadata
        self._number_of_split = number_of_splits
        self._classes_design = classes_design
        self._splits = []
        self._compute_splits(train_test_proportion, number_of_splits, pairing_column, selected_targets,
                             uniq_sample_id, balance_correction, classes_repartition, test_split_seed)

    def _compute_splits(self, train_test_proportion: float, number_of_splits: int, pairing_column: str,
                        selected_targets: List[str],  uniq_sample_id: List[str], 
                        balance_correction: int = 0,
                        classes_repartition: Union[dict, None] = None,
                        test_split_seed: Union[int,None] = None) -> None:
        """
        Create the desired number of split for the experiment. It includes/hadles the train-test repartition, the class
        balancing, the pairing of samples, the classes design, etc.

        Args:
            train_test_proportion (float): Proportion of the dataset to include in the test split.
            number_of_splits (int): total number of splits
            pairing_column (str): Column to user of the pairing. When empty ("") not pairing is done.
            selected_targets (List[str]): The selection of classes done with the interface or the automate.py (the names of the
                selected classes/targets)
                We consider selected_targets that has targets coming from multiple columns, that they are separated
                by "__" i.e. "ali__A", "med__B", etc.
            balance_correction (int, optional): balance correction to adjust proportion between classes.
                Defaults to 0. (for no balancing)
            classes_repartition (Union[dict, None], optional): . Defaults to None.
            test_split_seed (int | None, optional): Split seed number. For test purpose only,
                to be used from automate.py to test one specific split. Defaults to None.
        """
        self._logger.info("_compute_split function beginning")

        # 1 - filter out the samples having a target not included in the classification design
        # retrieve metadata dataframe
        df_filter = self._metadata.get_metadata()
        # keep only the lines for which the value in the final_targets column is in selected_targets
        df_filter = df_filter[df_filter[self._metadata.get_target_column()].isin(selected_targets)]
        # keep only the lines that correspond to samples in data file
        # (handles the cases of a metadata file for multiple data files : where samples in 
        # the metadata having corresponding targets are not in the provided data file)
        df_filter = df_filter[df_filter[self._metadata.get_id_column()].isin(uniq_sample_id)]

        # 2 - select only one sample per entity
        if pairing_column != "":
            # sort the dataframe by the pairing_column values
            df_entity = df_filter.sort_values(pairing_column)
            # group samples by the pairing column and keep only the first row of each group (.nth(0) is more stable
            # than .first())
            # Carefull : the groupby function change the index of the dataframe to the column it groups by
            df_entity = df_entity.groupby(pairing_column).nth(0)
        else:
            df_entity = df_filter

        # 2.5 - extract ids and targets, transform targets to labels
        ids = df_entity[self._metadata.get_id_column()]
        targets = df_entity[self._metadata.get_target_column()]
        labels = Utils.load_classes_from_targets(self._classes_design, targets)

        # 3- procede with the train-test division on the selected samples        
        if test_split_seed is not None:
            self._logger.debug(f"Testing split seed #{test_split_seed}")
            split_indexes: list[int] = [test_split_seed] # Test only one split seed
        else:
            split_indexes = list(range(number_of_splits)) # All splits indexes
            
        for split_index in split_indexes:
            if pairing_column == "":
                X_train, X_test, y_train, y_test = train_test_split(ids, labels, test_size=train_test_proportion,
                                                                    random_state=split_index, stratify=labels)

                # 4- retrieve the paired samples corresponding to the one in train or test set
            else:
                # random shuffle initialisation for second shuffle of samples
                rng = np.random.default_rng(seed=split_index)
                # define the ids column as the index of the dataframe, so it can be extracted with groupby().groups
                df = df_filter.set_index(self._metadata.get_id_column())
                # groups is a dictionary with 'keys' as the pairing value and 'values' as the index of the lines corresponding to the pairing
                groups = df.groupby(pairing_column).groups
                # apply the train-test division on the pairing values / the entity
                # TODO : careful check if labels is in the right order with the data
                X_train_temp, X_test_temp, y_train_temp, y_test_temp = train_test_split(df_entity.index, labels,
                                                                                        test_size=train_test_proportion,
                                                                                        random_state=split_index,
                                                                                        stratify=labels)
                # retrieve the ids corresponding the to entities in train
                X_train = []
                for representative in X_train_temp:
                    represented_pairing_value = df_filter.loc[representative][pairing_column]
                    X_train.extend(groups[represented_pairing_value])
                # retrieve targets corresponding to ids and then convert to labels
                X_train = pd.Series(X_train)
                targets = df.loc[X_train][self._metadata.get_target_column()]
                y_train = Utils.load_classes_from_targets(self._classes_design, targets)

                training_data = list(zip(X_train, y_train))
                rng.shuffle(training_data)
                X_train, y_train = zip(*training_data)

                # retrieve the ids corresponding the to entities in test
                X_test = []
                for representative in X_test_temp:
                    represented_pairing_value = df_filter.loc[representative][pairing_column]
                    X_test.extend(groups[represented_pairing_value])
                # retrieve targets corresponding to ids and then convert to labels
                X_test = pd.Series(X_test)
                targets = df.loc[X_test][self._metadata.get_target_column()]
                y_test = Utils.load_classes_from_targets(self._classes_design, targets)

                testing_data = list(zip(X_test, y_test))
                rng.shuffle(testing_data)
                X_test, y_test = zip(*testing_data)

            if balance_correction > 0:
                X_train, y_train = Utils.remove_random_samples_from_class(X_train,
                                                                          y_train,
                                                                          balance_correction,
                                                                          classes_repartition)
            X_train = list(X_train)
            y_train = list(y_train)
            X_test = list(X_test)
            y_test = list(y_test)
            
            if not self._validate_split(y_train, y_test):
                raise RuntimeError(f"_compute_split step #4 aborted for the invalid split #{split_index}.")
            
            self._splits.append([X_train, X_test, y_train, y_test])
            
        self._number_of_split = len(self._splits) # Update the number of splits if some have been removed
        
        self._logger.info("_compute_split function done")

    def load_split_with_index(self, split_index: int) -> list:
        return self._splits[split_index]

    def get_number_of_splits(self):
        """
        Return the attribute of number of split
        """
        return self._number_of_split

    def filter_sample_with_pairing_group(self, pairing_column: str) -> Tuple[List[str], List[str]]:
        """
        Function only needs the name of the column used to pair samples together.
        It retrieves other informations from the attributes (object MetaData).
        Then it iterates over all the metadata dataframe to store only one sample of each entity (entity stands for a
        biological source, like an individual). Multiple samples can originate from one entity.
        """
        metadata_dataframe = self._metadata.get_metadata()
        id_column = self._metadata.get_id_column()
        target_column = self._metadata.get_target_column()
        filtered_id = []
        filtered_target = []
        already_selected_value = set()
        # TODO : might want to change the process to sorting all lines and then picking the first one
        for index, row in metadata_dataframe.iterrows():
            if row[pairing_column] not in already_selected_value:
                already_selected_value.add(row[pairing_column])
                filtered_id.append(row[id_column])
                filtered_target.append(row[target_column])
        return filtered_id, filtered_target

    def get_selected_targets_and_ids(self, selected_targets: List[str], samples_id: List[str],
                                     targets: List[str]) -> Tuple[Tuple[str], Tuple[str]]:
        """
        Function just filters out the target/id that are not in the selected_targets list
        """
        return tuple(zip(*[(target, id) for target, id in zip(targets, samples_id) if target in selected_targets]))
    
    def _validate_split(self, y_train: list, y_test: list) -> bool:
                    # Test and train validation: they must have at least 2 classes.
        nb_test_classes: int = len(set(y_test))
        nb_train_classes: int = len(set(y_train))
        
        if nb_test_classes < 2 or nb_train_classes < 2:
            error_msg: str = "At least 2 classes must be present in both train and test splits."
            if nb_test_classes < 2:
                error_msg += f" Test set contains only the class '{next(iter(set(y_test)))}'."
            if nb_train_classes < 2:
                error_msg += f" Train set contains only the class '{next(iter(set(y_train)))}'."
            
            self._logger.error(error_msg)
            return False

        return True




from typing import Generator, Tuple, List, Dict, Union

import sklearn

from . import ClassificationDesign
from . import MetaData, MetaboModel
from .DataMatrix import DataMatrix
from .ExperimentDTO import ExperimentDTO
from .ModelFactory import ModelFactory
from ..conf.SupportedCV import CV_ALGORITHMS
from ..service import Utils
from ..service import init_logger

X_TRAIN_INDEX = 0
X_TEST_INDEX = 1
y_TRAIN_INDEX = 2
y_TEST_INDEX = 3

DEFAULT_NJOB = 1


class MetaboExperiment:
    def __init__(self):
        self._model_factory = ModelFactory()

        self._data_matrix = DataMatrix()
        self._is_progenesis_data = False
        self._metadata = MetaData()
        self._unique_ids = None

        self._number_of_splits = 25
        self._train_test_proportion = 0.2
        self._pairing_group_column = ""
        self._cv_folds = 5
        self._activate_multithreading = True

        self.classification_designs: Dict[str, ClassificationDesign] = {}

        self._supported_models = self._model_factory.create_supported_models("GS")
        self._custom_models = {}
        self._selected_models = []
        self._cv_algorithms = CV_ALGORITHMS
        self._selected_cv_type = list(self._cv_algorithms.keys())[0]
        
        self._logger = init_logger(__name__)

    def init_metadata(self):
        """
        Create an instance of MetaData for the attribute of metadata
        """
        self._metadata = MetaData()

    def get_metadata(self) -> MetaData:
        return self._metadata

    def get_final_targets_values(self):
        return self._metadata.get_final_targets_values()

    def set_final_targets_values(self, columns: List[str]):
        self._metadata.set_final_targets_values(columns)

    def add_final_targets_col_to_dataframe(self):
        self._metadata.add_final_targets_col_to_dataframe()

    def init_data_matrix(self):
        """
        Create an instance of DataMatrix for the attribute of data_matrix
        """
        self._data_matrix = DataMatrix()

    def set_metadata_with_dataframe(self, filename, data=None, from_base64=True):
        """
        Create an instance of MetaData and then fill it by reading and storing the content
        of the metadata file provided
        """
        self.init_metadata()
        self._metadata.read_format_and_store_metadata(filename, data=data, from_base64=from_base64)

    def set_data_matrix(self, path_data_matrix: str, data=None, from_base64: bool = True):
        """
        set metadata from progenesis data file while handling progenesis data file
        it creates a "fake" metadata structure from the retrieved id and classes of progenesis
        And formats it with the targets column
        """
        self._data_matrix.reset_file()
        # self._data_matrix handles itself to store data, and it also creates a "fake" metadata dataframe to keep
        # the same metadata format as if it was given by a metadata file
        # it returns None if it is not a progenesis file
        metadata_df, self._unique_ids = self._data_matrix.read_format_and_store_data(path_data_matrix, data=data, from_base64=from_base64)

        # format the metadata df (in case progenesis is given)
        self._is_progenesis_data = metadata_df is not None
        if metadata_df is not None:
            self._metadata = MetaData(metadata_df)
            self._metadata.set_id_column("sample_names")
            self._metadata.set_target_columns(["labels"])

    def get_data_matrix(self) -> DataMatrix:
        return self._data_matrix

    def get_train_test_proportion(self) -> float:
        """
        Retrieve the value of the attribute _train_test_proportion
        """
        return self._train_test_proportion

    def get_number_of_splits(self) -> int:
        """
        Retrieve the value of the attribute _number_of_splits
        """
        return self._number_of_splits

    def set_number_of_splits(self, number_of_splits: int):
        """
        Set the value of the MetaboExpe attribute _number_of_splits
        """
        self._number_of_splits = number_of_splits

    def set_train_test_proportion(self, train_test_proportion: float):
        """
        Set the value of the MetaboExpe attribute _train_test_proportion
        """
        self._train_test_proportion = train_test_proportion

    def create_splits(self, test_split_seed: int|None=None) -> None:
        """
        Check that Experiment parameters are set and then : create an instance of SplitGroup for each Classification Design
        (The init of SplitGroup triggers the _compute_splits function).
        If test_split_seed is provided, then only this test split seed is computed.
        """
        if self._number_of_splits is None:
            raise ValueError("Number of splits not set")
        if self._train_test_proportion is None:
            raise ValueError("Train test proportion not set")
        if self._pairing_group_column is None:
            raise ValueError("Pairing group column not set")
        if self._metadata is None:
            raise ValueError("Metadata not set")
        for _, classification_design in self.classification_designs.items():
            classification_design.set_split_parameter_and_compute_splits(self._train_test_proportion,
                                                                       self._number_of_splits, self._metadata,
                                                                       self._pairing_group_column, self._unique_ids, 
                                                                       test_split_seed)

    def get_pairing_group_column(self) -> str:
        """
        Retrieve the name of the column to use for pairing the samples
        """
        return self._pairing_group_column

    def set_pairing_group_column(self, pairing_group_column: str):
        """
        Set the value of the MetaboExpe attribute _pairing_group_column
        """
        if pairing_group_column not in self._metadata.get_columns():
            raise RuntimeError("Column {} is not in the metadata".format(pairing_group_column))
        self._pairing_group_column = pairing_group_column

    def get_classification_designs(self) -> Dict[str, ClassificationDesign]:
        """
        Retrieve all Classification Designs for an experience
        """
        return self.classification_designs

    def _raise_if_classes_design_is_not_valid(self, classes_design: dict) -> None:
        if "" in list(classes_design.keys()):
            raise ValueError("Empty label(s) is not allowed")
        if len(classes_design) < 2:
            raise ValueError("Labels must have different names")
        items = [inner_val for val in classes_design.values() for inner_val in val]
        if 0 in items:
            raise ValueError(
                "You need to select at least one class. If no class appears, "
                "please upload a data matrix and, if necessary, a metadata file"
            )
        if len(set(items)) != len(items):
            raise ValueError("Duplicate class name is not allowed")

    def add_classification_design(self, classes_design: dict):
        """
        add a Classification Design to the experience, creates an object ClassificationDesign each time
        classes_design: which target.s against which for the prediction, and the names (classes) of the groups of target.s
        """
        self._raise_if_classes_design_is_not_valid(classes_design)
        classification_design = ClassificationDesign(classes_design)
        self.classification_designs[classification_design.get_name()] = classification_design

    def remove_classification_design(self, name: str):
        """
        Remove a Classification Design, but not used
        """
        self.classification_designs.pop(name)

    def add_custom_model(self, model_name: str, needed_imports: str, params_grid: dict, importance_attribute: str):
        """
        Add the information needed to run a custom model
        model_name: the name of the model
        needed_imports: the modules needed to run this model
        params_grid: the hyperparameters grid with which hyperparameters to test and for which values
        importance_attribute: the attribute of the model to use to measure the importance of features
        """
        model_index = model_name.strip()
        model_index = "Custom" + model_index

        index = 1
        model_index = model_index + "0"
        # Check if model name already exists, if yes, get the next available index
        # This system make the addition of identical algorithm with different hyperparameters grid possible
        while model_index in self._custom_models:
            model_index = model_index[:-1]
            model_index = model_index + str(index)
            index += 1
        self._custom_models[model_index] = self._model_factory.create_custom_model(model_name, needed_imports,
                                                                                   params_grid, importance_attribute)

    def get_custom_models(self) -> dict:
        return self._custom_models

    def set_selected_models(self, selected_models: list):
        """
        Set the self._selected_models attribute with the list given in argument
        and for each Classification Design object initialize basics of Results instances
        selected_models: list of models to run during the experiment
        """
        if self.classification_designs == {}:
            raise ValueError("You must define at least one classification design before selecting models.")
        self._selected_models = selected_models
        for _, classification_design in self.classification_designs.items():
            classification_design.set_selected_models_name(selected_models)

    def update_classification_designs_with_selected_models(self):
        """

        """
        for _, classification_design in self.classification_designs.items():
            classification_design.set_selected_models_name(self._selected_models)

    def get_selected_models(self) -> list:
        """
        Retrieve the name of the selected models to use for the experiment
        """
        return self._selected_models

    def get_metadata_columns(self) -> list:
        """
        Retrieve the names of the columns in the metadata matrix
        """
        if self._metadata is None:
            raise RuntimeError("Metadata is not set.")
        return self._metadata.get_columns()

    def set_id_column(self, id_column: str):
        """
        Set the id_column attribute of the MetaData object
        It is the column in the metadata used to have a unique id for each line/item/sample
        id_column: string of the name of the column
        """
        if self._metadata is None:
            raise RuntimeError("Metadata is not set.")
        self._metadata.set_id_column(id_column)

    def validate_id_column(self) -> bool:
        """
        Ensure all values in the data index column (unique_ids) are in the the metadata id_column
        """
        if not self._is_progenesis_data and self._metadata is None:
            raise RuntimeError("Metadata is not set.")
        if self._unique_ids is None:
            return False
        return self._metadata.validate_id_column(self._is_progenesis_data, self._unique_ids)

    def get_unique_targets(self) -> list:
        """
        Retrieve a list of unique targets by applying a "set()" function to the complete list of targets.
        """
        try:
            return self._metadata.get_unique_targets()
        except RuntimeError:
            return []

    def get_model_from_name(self, model_name: str) -> MetaboModel:
        if model_name in self._supported_models.keys():
            return self._supported_models[model_name]
        elif model_name in self._custom_models.keys():
            return self._custom_models[model_name]
        else:
            raise RuntimeError(
                "The model '"
                + model_name
                + "' has not been found neither in supported and custom lists."
            )

    def _check_classification_design(self):
        error_message = "Train test proportion, number of splits and metadata need to be set before start learning: "
        if self._number_of_splits is None:
            raise RuntimeError(error_message + "missing number of splits")
        if self._train_test_proportion is None:
            raise RuntimeError(error_message + "missing train test proportion")
        if self._metadata is None:
            raise RuntimeError(error_message + "missing metadata")

    def all_classification_designs_names(self) -> Generator[Tuple[str, str], None, None]:
        """
        Retrieve all Classification Designs names for an experience.
        """
        for name, classification_design in self.classification_designs.items():
            yield name, classification_design.get_full_name()

    def reset_classification_designs(self):
        """
        Delete all existing Classification Designs.
        """
        self.classification_designs = {}

    def _raise_if_value_for_learning_not_setted(self):
        if self._data_matrix is None:
            raise RuntimeError("Data matrix not set")
        if self._metadata is None:
            raise RuntimeError("Metadata not set")
        if self._pairing_group_column is None:
            raise RuntimeError("Pairing group column not set")
        if self._train_test_proportion is None:
            raise RuntimeError("Train test proportion not set")
        if self._number_of_splits is None:
            raise RuntimeError("Number of splits not set")
        if self._selected_models is None:
            raise RuntimeError("Selected models not set")
        if self._cv_folds is None:
            raise RuntimeError("CV folds not set")
        if self.classification_designs == {}:
            raise RuntimeError(
                "You must define at least one classification design before learning."
            )

    def learn(self):
        self._raise_if_value_for_learning_not_setted()
        cv_algorithm_constructor = self.get_cv_algorithm_constructor()
        cv_algorithm_config = self.get_cv_algorithm_configuration()
        self._check_classification_design()
        self._data_matrix.load_data()
        assert(self._data_matrix.data is not None)
        if not self._data_matrix.data.index.isin(self._metadata.get_metadata()[self._metadata.get_id_column()]).all():
            raise ValueError(f"'{self._metadata.get_id_column()}' seen not to be as valid name of the unique id column.")
        params = []
        for _, classification_design in self.classification_designs.items():
            self._logger.info("-> Classification design loop ")
            selected_targets_name = classification_design.get_selected_targets_name()
            (selected_targets, selected_ids,) = self._metadata.get_selected_targets_and_ids(selected_targets_name)
            classes = Utils.load_classes_from_targets(
                classification_design.get_classes_design(), selected_targets
            )
            for split_index, split in classification_design.all_splits():
                for model_name in self._selected_models:
                    x_train = self._data_matrix.load_samples_corresponding_to_IDs_in_splits(
                        split[X_TRAIN_INDEX]
                    )
                    x_test = self._data_matrix.load_samples_corresponding_to_IDs_in_splits(
                        split[X_TEST_INDEX]
                    )
                    params.append(
                        (model_name, classification_design.get_name(), split_index, split, x_train, x_test,
                         cv_algorithm_constructor, cv_algorithm_config, selected_ids, classes)
                    )
        self.run_learning(params)
        self._data_matrix.unload_data()


    def run_learning(self, params: List[tuple]):
        self._logger.info("-> run_learning function beginning")
        # launch the run_on_model function with the params
        result_params = [self.run_on_model(*param) for param in params]

        for result_param in result_params:
            (
                classification_design_name, 
                model_name, 
                best_model, 
                scaled_data, 
                importance_attribute, 
                classes, 
                y_train,
                y_train_pred,
                y_test,
                y_test_pred,
                split_index,
                X_train,
                X_test
            ) = result_param
            results = self.classification_designs[classification_design_name].get_results()
            results[model_name].set_feature_names(X_train) # called multiple times but it's ok
            results[model_name].design_name = classification_design_name
            results[model_name].add_results_from_one_algo_on_one_split(
                best_model, scaled_data, importance_attribute, 
                classes, y_train, y_train_pred, y_test, y_test_pred,
                split_index, X_train.index, X_test.index
            )

        for classification_design_name, classification_design in self.classification_designs.items():
            results = classification_design.get_results()
            _, selected_sample_id = self._metadata.get_selected_targets_and_ids(
                classification_design.get_selected_targets_name())
            for model_name, result in results.items():
                self._logger.info(f"-> Compute remaining results for {model_name} of {classification_design_name} design")
                result.compute_remaining_results_on_all_splits(selected_sample_id)
            classification_design.set_is_done(True)


    def run_on_model(self, model_name, classification_design_name, split_index, split, x_train, x_test,
                     cv_algorithm_constructor, cv_algorithm_config, selected_ids, classes):
        self._logger.info(f"-> Split : {split_index}")
        self._logger.info(f"-> Model : {model_name}")

        metabo_model = self.get_model_from_name(model_name)

        # When set n_processes is set to -1, all processors are used.
        n_processes = -1 if self._activate_multithreading else DEFAULT_NJOB
        best_model = metabo_model.train(
            self._cv_folds,
            x_train,
            split[y_TRAIN_INDEX],
            cv_algorithm_constructor,
            cv_algorithm_config,
            n_processes,
            seed=split_index
        )
        y_train_pred = best_model.predict(x_train)
        y_test_pred = best_model.predict(x_test)
        return (
            classification_design_name,
            model_name,
            best_model,
            self._data_matrix.get_scaled_data(selected_ids),
            metabo_model.get_importance_attribute(),
            classes,
            split[y_TRAIN_INDEX],
            y_train_pred,
            split[y_TEST_INDEX],
            y_test_pred,
            str(split_index),
            x_train,
            x_test
        )

    def get_results(self, classes_design: str, algo_name) -> dict:
        return self.classification_designs[classes_design].get_results()[algo_name]

    def get_all_updated_results(self) -> dict:
        """
        Retrieve, for each Classification Design that is done, the results dict corresponding
        """
        results = {}
        for name in self.classification_designs:
            if self.classification_designs[name].get_is_done():
                results[name] = self.classification_designs[name].get_results()
        return results

    def get_all_algos_names(self) -> list:
        """
        Concatenate the list of names from default (supported) models and custom models
        """
        return list(self._supported_models.keys()) + list(self._custom_models.keys())

    def set_cv_type(self, cv_type: str):
        """
        Set the type of Cross-Validation (cv) for the experiment
        """
        if cv_type not in self._cv_algorithms:
            raise ValueError(f"CV type '{cv_type}' is not supported. Choices are : {list(self._cv_algorithms.keys())}")

        if cv_type == "GridSearchCV":
            self._supported_models = self._model_factory.create_supported_models("GS")
        elif cv_type == "RandomizedSearchCV":
            self._supported_models = self._model_factory.create_supported_models("RS")

        self._selected_cv_type = cv_type

    def get_selected_cv_type(self) -> str:
        """
        Return the type of CV selected for this experiment
        """
        return self._selected_cv_type

    def get_cv_algorithm_constructor(self) -> sklearn.model_selection:
        return self._cv_algorithms[self._selected_cv_type]["constructor"]

    def get_cv_algorithm_configuration(self) -> list:
        return self._cv_algorithms[self._selected_cv_type]["params"]

    def set_cv_algorithm_configuration(self, cv_algorithm_configuration: list):
        config_index = 0
        for param_index, params in enumerate(self._cv_algorithms[self._selected_cv_type]["params"]):
            if not params["constant"]:
                self._cv_algorithms[self._selected_cv_type]["params"][param_index]["value"] = \
                    cv_algorithm_configuration[config_index]
                config_index += 1

    def get_cv_types(self) -> List[str]:
        """
        Retrieve the dict of possible cross-validation types supported by the PICO
        """
        return list(self._cv_algorithms.keys())

    def generate_save(self) -> ExperimentDTO:
        """
        Create an object ExperimentDTO
        (which is a holder of some MetaboExperiment attributes)
        """
        return ExperimentDTO(self)

    def full_restore(self, saved_metabo_experiment_dto: ExperimentDTO):
        """
        Restore an experiment from a saving (always? from file)
        Data and parameters
        """
        self._metadata = saved_metabo_experiment_dto.metadata
        self._data_matrix = saved_metabo_experiment_dto.data_matrix
        self._static_restore_for_partial(saved_metabo_experiment_dto)

    def _static_restore_for_partial(self, saved_metabo_experiment_dto: ExperimentDTO):
        """
        ??? STATIC vs partial_restore ???
        Restore attributes from a ExperimentDTO object
        a "Partial" restore means it only brings back parameters and no data
        """
        self._number_of_splits = saved_metabo_experiment_dto.number_of_splits
        self._train_test_proportion = saved_metabo_experiment_dto.train_test_proportion
        self.classification_designs = saved_metabo_experiment_dto.classification_designs
        self._custom_models = saved_metabo_experiment_dto.custom_models
        self._selected_models = saved_metabo_experiment_dto.selected_models
        self._selected_cv_type = saved_metabo_experiment_dto.selected_cv_type

    def partial_restore(self, saved_metabo_experiment_dto: ExperimentDTO, filename_data: str, filename_metadata: str,
                        data=None, from_base64_data: bool = True, metadata=None, from_base64_metadata=True,):
        """
        ??? STATIC vs partial_restore ???
        Do a partial restore of an experiment
        """
        self._data_matrix.set_raw_use(saved_metabo_experiment_dto.data_matrix.is_raw())
        self._data_matrix.set_remove_rt(saved_metabo_experiment_dto.data_matrix.get_remove_rt())
        self.set_data_matrix(filename_data, data=data, from_base64=from_base64_data,)
        self.set_metadata_with_dataframe(filename_metadata, data=metadata, from_base64=from_base64_metadata)
        self._static_restore_for_partial(saved_metabo_experiment_dto)

    def load_results(self, saved_metabo_experiment_dto: ExperimentDTO):
        """
        Init a new experiment (new metadata and data_matrix) and load saved results
        """
        self.init_metadata()
        self.init_data_matrix()
        self._static_restore_for_partial(saved_metabo_experiment_dto)

    def is_save_safe(self, saved_metabo_experiment_dto: ExperimentDTO) -> bool:
        """
        Verify that the hash from the saved ExperimentDTO is the same from the current object
        """
        return (self._metadata.get_hash() == saved_metabo_experiment_dto.metadata.get_hash()
                and self._data_matrix.get_hash() == saved_metabo_experiment_dto.data_matrix.get_hash())

    def is_the_data_matrix_corresponding(self, data: str) -> bool:
        return self._data_matrix.get_hash() == Utils.compute_hash(data)

    def is_the_metadata_corresponding(self, metadata: str) -> bool:
        return self._metadata.get_hash() == Utils.compute_hash(metadata)

    def get_target_column(self) -> str:
        """
        Retrieve the _target_column attribute of metadata
        """
        return self._metadata.get_target_column()

    def get_id_column(self) -> str:
        """
        Retrieve the _id_column attribute of metadata
        """
        return self._metadata.get_id_column()

    def is_progenesis_data(self) -> bool:
        """
        Return the bool indicating if the data given is of the progenesis format
        """
        return self._is_progenesis_data

    def is_data_raw(self) -> Union[bool, None]:
        """
        return the bool indicating to use either raw or normalized data
        """
        return self._data_matrix.is_raw()

    def set_raw_use_for_data(self, use_raw: bool):
        """
        Set the bool value of whether to use the raw data from a progenesis matrix or the normalized data
        """
        self._data_matrix.set_raw_use(use_raw)

    def get_data_matrix_remove_rt(self) -> bool:
        """
        return the value of the _remove_rt attribute
        (if true, remove the features detected before 1 minute of Retention Time)
        """
        return self._data_matrix.get_remove_rt()

    def set_data_matrix_remove_rt(self, remove_rt: bool):
        """
        set the value of the _remove_rt attribute
        (if true, remove the features detected before 1 minute of Retention Time)
        """
        self._data_matrix.set_remove_rt(remove_rt)

    def get_cv_folds(self) -> int:
        """
        Return the number of Cross Validation folds
        """
        return self._cv_folds

    def set_cv_folds(self, cv_folds: int):
        """
        Set the number of Cross Validation folds
        """
        if cv_folds < 2:
            raise ValueError("CV folds must be greater than or equal to 2.")
        self._cv_folds = cv_folds

    def data_is_set(self) -> bool:
        return self._data_matrix.data_is_set()

    def metadata_is_set(self) -> bool:
        return self._metadata.metadata_is_set()

    def set_target_columns(self, target_cols: List[str]) -> None:
        self._metadata.set_target_columns(target_cols)

    def set_multithreading(self, activate_multithreading: bool):
        self._activate_multithreading = activate_multithreading

    def get_samples_id(self):
        """
        This used to return self._metadata.get_samples_id(), thus sourcing samples name from metadata file.
        It has been adjusted to account for the situation of metadata containing more information
        than the samples studied by the provided data file
        """
        if self._unique_ids is None:
            return []
        return self._unique_ids
    
    def get_classes_repartition_for_all_experiment(self) -> dict:
        classes_repartition = {}
        for classification_design_name in self.classification_designs:
            class_design = self.classification_designs[classification_design_name].get_classes_design()
            classes_repartition[classification_design_name] = \
                self._metadata.get_classes_repartition_based_on_design(class_design)
        return classes_repartition

    def get_balance_correction_for_all_experiment(self) -> dict:
        balance_correction = {}
        for classification_design_name in self.classification_designs:
            balance_correction[classification_design_name] = \
                self.classification_designs[classification_design_name].get_balance_correction()
        return balance_correction

    def set_balance_correction_for_experiment(self, classification_design_name: str, balance_correction: int) -> None:
        if classification_design_name not in self.classification_designs:
            raise ValueError("Classification design name not found")
        self.classification_designs[classification_design_name].set_balance_correction(balance_correction)

    def display_splits(self) -> None:
        """
        Display the classes repartition for each split of each Classification Design.
        See more détail in the Controller.display_splits() method description.
        """
        from collections import Counter

        def display_classes_repartition(target_classes: list) -> str:
            return ' vs '.join([f"{cnt} ({int(round(cnt*100 / len(target_classes))):02d}%)" for _, cnt in sorted(Counter(target_classes).items())])

        for key,classification_design in self.classification_designs.items():
            balance_corr: int = classification_design.get_balance_correction()
            experimental_classes_repartition: dict = self._metadata.get_classes_repartition_based_on_design(classification_design.get_classes_design())
            total_cnt: int = sum(value for value in experimental_classes_repartition.values())

            debug_lines = []
            for split_index, split_group in classification_design.all_splits():
                if split_index == 0:
                    class_set = sorted(set(split_group[2]))
                    debug_lines.append(f"Classification Design '{key}' details:")
                    total_class_repartition: str = " vs ".join([f"'{cl}': {cnt:02d} ({int(round(cnt*100 / total_cnt)):02d}%)"
                                                                for cl, cnt in sorted(experimental_classes_repartition.items())])
                    debug_lines.append(f"Data set repartition: {total_class_repartition} (Balance corr={balance_corr}%).")
                    debug_lines.append(f"Classes '{class_set[0]}' vs '{class_set[1]}' repartition in splits (All | Train | Test):")

                all_cls_cnt: str = display_classes_repartition(split_group[2] + split_group[3])
                train_cls_cnt: str = display_classes_repartition(split_group[2])
                test_cls_cnt: str = display_classes_repartition(split_group[3])
                debug_lines.append(f"Split #{split_index:02d}: All=[{all_cls_cnt}] | Train=[{train_cls_cnt}] | Test=[{test_cls_cnt}]")

            debug_message = "\n\t".join(debug_lines)
            self._logger.debug(debug_message)


from .DataFormat import DataFormat

from .RunMLalgo import runAlgo
from .SamplesPairing import SamplesPairing
from .Utils import *
from .Plots import Plots
from .LoggerConfig import set_log_filename, init_logger, log_exceptions



import logging
import coloredlogs # type: ignore
from datetime import datetime
import inspect
import os
import threading
import traceback
from functools import wraps
from typing import Callable
from .Utils import get_pico_subdir

log_filename: str|None = None  # Global variable for log filename

def log_exceptions(logger: logging.Logger) -> Callable:
    def decorator(func: Callable):
        @wraps(func)
        def wrapper(*args, **kwargs) -> Callable:
            try:
                return func(*args, **kwargs)
            except Exception as e:
                # Log the exception or handle it as needed
                thread_name = threading.current_thread().name
                logger.error(f"Error in thread {thread_name}: {e}\n{traceback.format_exc()}")
                raise # Re-raise the exception to preserve the original
        return wrapper
    return decorator


def set_log_filename(filename: str="pico.log", add_date: bool=True, level=logging.DEBUG) -> logging.Logger:
    """Sets the log filename with an optional date suffix.
    Args:
        filename (str, optional): The base filename for the log. Defaults to "pico.log".
        add_date (bool, optional): If True, adds the current date to the filename. Defaults to True.
    """
    global log_filename

    if add_date:
        date_suffix = datetime.now().strftime("%Y-%m-%d")
        
        if filename.lower().endswith(".log"):
            log_filename = filename.replace(".log", f"_{date_suffix}.log")
        else:
            log_filename = f"{filename}_{date_suffix}.log"
    else:
        log_filename = filename

    # Log file in ~/pico_files/logs directory
    logs_directory = get_pico_subdir("logs")
    log_filename = os.path.join(logs_directory, log_filename)      

    # Add "-----------------------" in the log file to start the current session
    with open(log_filename, 'a') as log_file:
        if threading.current_thread() is threading.main_thread():
            log_file.write(f"\n{'----- New start (' + threading.current_thread().name + ') ':-<80}\n")
        else:
            log_file.write(f"      New thread ({threading.current_thread().name})\n")

    # Root logger
    root_logger = logging.getLogger()
    root_logger.handlers = []  # Clear existing handlers
    root_logger.setLevel(logging.WARNING)  # Only show WARNING level and above
    root_logger.propagate = False

    # Terminal (console) handler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.WARNING)

    # File handler
    file_handler = logging.FileHandler(log_filename)
    file_handler.setLevel(logging.DEBUG)  # Ensure DEBUG level for detailed logs

    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    console_handler.setFormatter(formatter)
    file_handler.setFormatter(formatter)

    root_logger.addHandler(console_handler)
    root_logger.addHandler(file_handler)

    # Ensure colored logs for terminal
    coloredlogs.install(level=logging.WARNING, logger=root_logger, stream=console_handler.stream)

    # werkzeug logs
    werkzeug_logger = logging.getLogger('werkzeug')
    werkzeug_logger.setLevel(logging.INFO)  # Log level to INFO for console
    
    # Add logs for werkzeug in terminal
    werkzeug_console_handler = logging.StreamHandler()
    werkzeug_console_handler.setLevel(logging.INFO)
    werkzeug_console_handler.setFormatter(formatter)
    werkzeug_logger.addHandler(werkzeug_console_handler)
    werkzeug_logger.propagate = False

    return  init_logger()


def init_logger(module_name: str|None=None, level=logging.DEBUG) -> logging.Logger:
    """"pico.log"

    Args:
        module_name (str | None, optional): The name of the module for the logger. Defaults to None.
        level (int, optional): The logging level. Defaults to logging.DEBUG.
            Levels (from high to low): logging.CRITICAL logging.ERROR
                                       logging.WARNING logging.INFO logging.DEBUG

    Returns:
        logging.Logger: The configured logger instance
    """
    
    global log_filename
                   
    def get_module_name() -> str | None:
        cur_frame = inspect.currentframe()
        if not cur_frame:
            return None
        frame = cur_frame.f_back
        if not frame:
            return None
        module = inspect.getmodule(frame)
        if not module:
            return None
        return module.__name__

    if module_name is None:
        module_name = get_module_name() or "pico"
            
    logger = logging.getLogger(module_name)
    logger.setLevel(level)
    logger.propagate = False
    
    if not logger.handlers:
        # Terminal (console) handler
        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levellevel)s - %(message)s')
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)        
        
        # File handler
        if log_filename is not None:
            file_handler = logging.FileHandler(log_filename)
            file_handler.setLevel(level)
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            file_handler.setFormatter(formatter)
            
            logger.addHandler(file_handler)

        # Ensure colored logs for terminal
        coloredlogs.install(level=level, logger=logger, stream=console_handler.stream)

    return logger



import pandas as pd
import sklearn
from sklearn.model_selection import RandomizedSearchCV
from ..service import init_logger, log_exceptions

class MetaboModel:
    def __init__(self, model: sklearn, grid_search_configuration: dict, importance_attribute: str):
        self._logger = init_logger()
        self.train = log_exceptions(self._logger)(self.train)  # Apply log_exception decorator
        self.grid_search_param = grid_search_configuration
        self.model = model

        self.importance_attribute = importance_attribute

    def train(
        self,
        folds: int,
        X_train: pd.DataFrame,
        y_train: list,
        cv_algorithms: sklearn.model_selection,
        cv_algorithm_config,
        number_of_processes: int, seed: int
    ) -> sklearn:
        if cv_algorithms == RandomizedSearchCV:
            search = cv_algorithms(
                self.model(random_state=seed),
                self.grid_search_param,
                cv=folds,
                random_state=seed,
                n_jobs=number_of_processes,
                n_iter=cv_algorithm_config[0]["value"],
            )
        else:
            search = cv_algorithms(
                self.model(random_state=seed),
                self.grid_search_param,
                cv=folds,
                n_jobs=number_of_processes,
            )
        
        try:
            search.fit(X_train, y_train)
        except Exception as e:
            message: str = f"{cv_algorithms = } cv{folds:d} folds | X_train size {len(X_train)}) y_train length = {len(y_train)}"
            self._logger.error(message)
            raise
        
        return search.best_estimator_

    def get_importance_attribute(self):
        return self.importance_attribute

# TODO: dump best model ?



import base64
import io
import os.path
from typing import List, Tuple, Dict

import pandas as pd

from ..service import compute_hash
from ..service import init_logger

ROOT_PATH = os.path.dirname(__file__)
DUMP_PATH = os.path.join(ROOT_PATH, os.path.join("dumps", "metadata"))
DUMP_METADATA_PATH = os.path.join(DUMP_PATH, "metadata.p")
DUMP_METADATA_COLUMNS_PATH = os.path.join(DUMP_PATH, "metadata_columns.p")
DUMP_SAMPLES_ID_PATH = os.path.join(DUMP_PATH, "samples_id.p")
DUMP_TARGETS_PATH = os.path.join(DUMP_PATH, "targets.p")


class MetaData:
    def __init__(self, metadata_dataframe: pd.DataFrame = None):
        self._logger = init_logger()
        self._dataframe = metadata_dataframe

        self._id_column = None
        self._target_column = None
        self._final_targets_values = []

        self._hash = None

    def get_final_targets_values(self):
        """
        get the value of the attribute
        """
        return self._final_targets_values

    def set_final_targets_values(self, columns: List[str]):
        """
        compute the list of new strings (unite targets str with '__') and
        modify the value of the attribute with the list of values
        """
        if len(columns) > 1:
            df = self._dataframe.loc[:, columns].astype(str)
            final = list(df.apply(lambda row: "__".join(row), axis=1))
        else:
            final = self._dataframe[columns[0]].tolist()
        self._final_targets_values = final

    def add_final_targets_col_to_dataframe(self):
        """
        modify the attribute _dataframe to add the column of unified targets values
        """
        self._dataframe["final_targets"] = self._final_targets_values

    def add_final_targets_col_to_dump(self):
        """
        modify the temporary save of metadata (dump)
        """
        return None

    def read_format_and_store_metadata(self, path, data=None, from_base64=True):
        """
        Read the file to get a dataframe and compute a hash on it
        Store de obtained metadata dataframe in the attribute
        """
        df = self._load_and_format(path, data=data, from_base64=from_base64)
        if data is not None:
            self._hash = compute_hash(data)
        self._dataframe = df

    def get_hash(self) -> str:
        return self._hash

    def _load_and_format(self, filename, data=None, from_base64=True) -> pd.DataFrame:
        """
        Read the file and decode it to put it in the pd.Dataframe format
        """
        if from_base64:
            data_type, data_string = data.split(",")
            data = base64.b64decode(data_string)
            # print("data decoded :{}")
            # print(data[:200])
        else:
            data = filename

        if "csv" in filename:
            # Assume that the user uploaded a CSV file
            if from_base64:
                data = io.StringIO(data.decode("utf-8"))
            df = pd.read_csv(data, sep=None, na_filter=False, engine="python")
        elif "xls" in filename:
            if from_base64:
                data = io.BytesIO(data)
            # Assume that the user uploaded an excel file
            df = pd.read_excel(data, na_filter=False)
        else:
            raise TypeError("The input file is not of the right type, must be excel or csv.")
        return df

    def get_metadata(self) -> pd.DataFrame:
        """
        Getter for the metadata DataFrame
        """
        if self._dataframe is None:
            raise RuntimeError("Try to access the metadata before setting it.")
        return self._dataframe

    def get_columns(self) -> List[str]:
        """
        Retrieve the names of the columns in the metadata matrix as a list
        """
        if self._dataframe is None:
            self._logger.info("MetaData get_columns self._dataframe is None")
            return []
        return self._dataframe.columns.tolist()

    def get_unique_targets(self) -> List[str]:
        """
        Retrieve a list of unique targets by applying a "set()" function to the complete list of targets.
        eg: [0, 1, 1, 0, 0, 0, 1, 0, 2] will return [0, 1, 2]
        """
        targets = self.get_targets()
        return list(set(targets))

    def set_id_column(self, id_column: str) -> None:
        """
        Set the id_column attribute of the MetaData object
        It is the column in the metadata used to have a unique id for each line/item/sample
        id_column: string of the name of the column
        """
        if id_column not in self.get_columns():
            raise ValueError(f"'{id_column}' is not a column of the metadata. The columns are: {self.get_columns()}")
        self._id_column = id_column

    def validate_id_column(self, is_progenesis_data: bool, unique_ids: List[str]) -> bool:
        """
        Ensure all values in the data index column (unique_ids) are in the the metadata id_column
        """
        df = self._dataframe
        if self._dataframe is None:
            return False
        try:
            ids = df[self._id_column]
        except KeyError as e:
            self._logger.error(f"Error: column '{self._id_column}' does not exist")
            return False
        except TypeError as e:
            self._logger.error(f"Error: values in column '{self._id_column}' are not comparable")
            return False
        
        if is_progenesis_data:
            return ids.isin(unique_ids).all()
        else:
            # 'ids' represents the sample names in the metadata
            # 'unique_ids' represents sample names in the data
            # some experiments could have one metadata file for multiple data files, so we should check 
            # if the unique_ids of the data file are all contained in the metadata file.
            return set(ids) >= set(unique_ids)

    def set_target_column(self, target_column: List[str]) -> None:
        """
        Define which of the metadata columns is the targets column
        """
        if target_column not in self.get_columns():
            raise ValueError(f"'{target_column}' is not a column of the metadata. The columns are: {self.get_columns()}")
        self._target_column = target_column

    def get_target_column(self) -> List[str]:
        """
        Provide the _target_column attribute
        """
        return self._target_column

    def get_id_column(self) -> str:
        """
        Provide the _id_column attribute
        """
        return self._id_column

    def get_targets(self) -> List[str]:
        """
        return a list of all the targets
        either just the targets indicated in the column
        or "__".join() of the multiple target column selected
        """
        if self._target_column is None:
            self._logger.warning("Accessing targets before setting the column")
            return []
        return self._dataframe.loc[:, self._target_column]

    def get_selected_targets_and_ids(self, selected_targets: List[str]) -> Tuple[Tuple[str], Tuple[str]]:
        return tuple(zip(*[(target, id) for target, id in zip(self.get_targets(), self.get_samples_id()) if target in selected_targets]))

    def get_selected_targets(self, selected_targets: List[str]) -> List[str]:
        if selected_targets is None or len(selected_targets) == 0:
            raise ValueError("No target selected")
        return [target for target in self.get_targets() if target in selected_targets]

    def get_samples_id(self) -> List[str]:
        if self._id_column is None:
            self._logger.warning("Accessing samples id before setting the column")
            return []
        return self._dataframe[self._id_column].tolist()

    def metadata_is_set(self) -> bool:
        return self._dataframe is not None

    def set_target_columns(self, target_cols: List[str]) -> None:
        """
        Create the new targets and add them to the metadata dataframe
        """
        # Create the new targets from the given metadata informations (NAME__NAME)
        self.set_final_targets_values(target_cols)
        # Add the values of the final (new) targets to the dataframe of metadata (in memory)
        self.add_final_targets_col_to_dataframe()
        # Define the name of the column of targets as final_targets
        self.set_target_column("final_targets")

    def get_classes_repartition_based_on_design(self, classes_design: Dict[str, List[str]]) -> Dict[str, int]:
        """
        Retrieve the number of target for each class
        Which can be used to display the ratio of each classes in the experiment/design
        """
        target_repartition = self._dataframe["final_targets"].value_counts().to_dict()
        classes_repartition = {}
        for class_name in classes_design:
            classes_repartition[class_name] = sum([target_repartition[target] for target in classes_design[class_name]])
        return classes_repartition

# TODO: join sampleId and target in same pickle file



from sklearn.model_selection import train_test_split
import glob, random
import numpy as np


class SamplesPairing:
    def __init__(
        self, pairings, sample_names, targets, IDs, proportion_in_test, nbr_splits
    ):
        self.pairings = pairings
        self.sample_names = sample_names
        self.targets = targets
        self.ids = IDs
        self.proportion = float(proportion_in_test)
        self.nbr_splits = int(nbr_splits)
        self.dict_splits = {}

        self.names_dict = {n: idx for idx, n in enumerate(self.sample_names)}

    def split(self):
        """
        We assume there can be maximum two types of pairing simultaneously
        :return: Nothing
        """
        if not self._is_there_pairing_to_do():

            ###### Creation de X et y ######

            X = []
            y = []
            for s in self.sample_names:  # itère sur chq nom de sample
                for i, id in enumerate(self.ids):  # itère sur chq id unique
                    if id in s:  # vérifie si l'id est présent dans le nom de sample
                        X.append(s)
                        y.append(
                            self.targets[i]
                        )  # ajoute label correspondant à l'id(correspondant lui-meme au sample)

            ##### Create the splits ######

            for i in range(self.nbr_splits):
                X_train, X_test, y_train, y_test = train_test_split(
                    X, y, test_size=self.proportion, random_state=i, stratify=y
                )
                # convert sample names to indices so its easier to reconstruct later
                # X_train = [self.names_dict[i] for i in X_train]
                # X_test = [self.names_dict[i] for i in X_test]

                ###### save splits dans dict ######

                self.dict_splits["split{}".format(i)] = [
                    X_train,
                    X_test,
                    y_train,
                    y_test,
                ]
        else:
            self._test_pairing_patterns_spelling()
            # select base files to do the spliting on

            ###### Creation de X et y ######

            X = []
            y = []
            first_patterns = [i[0] for i in self.pairings]
            for s in self.sample_names:  # itère sur chaque nom de sample
                ok = True
                for p in first_patterns:  # itère sur chq 1er pattern des pairings
                    if (
                        p not in s
                    ):  # check si le (ou un des deux) pattern n'est pas présent
                        ok = False
                # if it does, check its ID to match its target
                if (
                    ok
                ):  # si le/les patterns sont présents, ajoute le sample au groupe de base/de reference
                    for i, id in enumerate(self.ids):
                        if id in s:
                            X.append(s)
                            y.append(self.targets[i])

            ##### Create the splits ######

            for i in range(self.nbr_splits):
                X_train, X_test, y_train, y_train = train_test_split(
                    X, y, test_size=self.proportion, random_state=i, stratify=y
                )

                # replace pattern to match spliting
                # create target list accordingly

                pattern_list = first_patterns
                idx = len(self.pairings) - 1
                new_xy = []
                if idx + 1 == 1:  # if there is one pairing to do
                    new_xy = self._iterate_on_pattern_to_get_paired_files(
                        pattern_list, idx, self.pairings[idx], X_train, X_test
                    )
                elif idx + 1 == 2:  # if there is two pairings to do
                    X1 = X_train
                    X2 = X_test
                    for p in self.pairings[0]:
                        pattern_list[0] = p
                        new_sub_xy = self._iterate_on_pattern_to_get_paired_files(
                            pattern_list, idx, self.pairings[idx], X1, X2
                        )
                        X1 = new_sub_xy[0][0]
                        X2 = new_sub_xy[0][1]

                        new_xy.extend(
                            new_sub_xy
                        )  # in shape of a list containing all [xtrain, xtest, ytrain, ytest] groups

                ###### save splits dans dict ######

                # shuffles
                new_xy = np.swapaxes(
                    new_xy, 0, 1
                )  # is now in shape of 4 big lists (xtrain, xtest, ytrain, ytest)

                new_X_train = new_xy[0]
                new_y_train = new_xy[2]
                train_zip = list(zip(new_X_train, new_y_train))
                random.Random(13).shuffle(train_zip)
                X_train, y_train = zip(*train_zip)

                new_X_test = new_xy[1]
                new_y_test = new_xy[3]
                test_zip = list(zip(new_X_test, new_y_test))
                random.Random(13).shuffle(test_zip)
                X_test, y_test = zip(*test_zip)

                # save to dict
                X_train = [self.names_dict[i] for i in X_train]
                X_test = [self.names_dict[i] for i in X_test]

                self.dict_splits["split{}".format(i)] = [
                    X_train,
                    X_test,
                    y_train,
                    y_test,
                ]

    def _is_there_pairing_to_do(self):
        no_pairing = 0
        for p in self.pairings:
            if p == "no":
                no_pairing += 1
        if no_pairing == len(self.pairings):
            return False
        else:
            return True

    def _iterate_on_pattern_to_get_paired_files(self, base, idx, pairing, X1, X2):
        """

        :param base: list being updated of the patterns
        :param idx: index of the pairing to fit with base shape
        :param pairing: list of patterns for a single pairing
        :param X1: Xtrain data to replace pattern in
        :param X2: Xtest data to replace pattern in
        :return: lists of new xtrain, xtest, ytrain and ytest
        """
        p1 = base[idx]
        new_xy = []
        for pattern in pairing:
            base[idx] = pattern
            X1_2, X2_2, y1_2, y2_2 = self._replace_pattern_get_matching_files(
                p1, base[idx], X1, X2
            )
            new_xy.append([X1_2, X2_2, y1_2, y2_2])

        return new_xy

    def _replace_pattern_get_matching_files(self, p1, p2, Xtrain, Xtest):
        Xtrain_2 = [i.replace(p1, p2) for i in Xtrain]
        Xtest_2 = [i.replace(p1, p2) for i in Xtest]
        ytrain_2 = []
        ytest_2 = []

        for file in Xtrain_2:
            for i, id in enumerate(self.ids):
                if id in file.split("/")[-1]:
                    ytrain_2.append(self.targets[i])
        for file in Xtest_2:
            for i, id in enumerate(self.ids):
                if id in file.split("/")[-1]:
                    ytest_2.append(self.targets[i])

        return Xtrain_2, Xtest_2, ytrain_2, ytest_2

    def _test_pairing_patterns_spelling(self):
        """
        If the list obtained directly from the directory with a glob.glob is different of the list obtained by replacing
        the pattern from the "base list" (list of file containing the first pattern of the pairing) it means there is a
        problem of pattern spelling that will lead to further trouble in the rest of the processing
        :return: nothing
        """
        path_of_files = self.files.split("/")[:-1]
        path_of_files = "/".join(path_of_files)

        prob_patterns = []

        # TODO
        # not sure if all the cases are covered, especially the case where the pattern is not found and return the
        # same list unchanged
        for pairing in self.pairings:
            p1 = glob.glob(path_of_files + "*" + pairing[0] + "*")
            for pattern in pairing:
                p = glob.glob(path_of_files + "*" + pattern + "*")
                p_by_replace = [f.replace(pairing[0], pattern) for f in p1]

                if p.sort() != p_by_replace.sort():
                    prob_patterns.append(pattern)

        if len(prob_patterns) > 0:
            raise (
                "There is a pattern error, here are the problematic patterns : {}".format(
                    prob_patterns
                )
            )


# #Conditionnal statement to handle split creation with pairing(s)
# if pairing_pn == "no" and pairing_12 == "no":  # no pairing
#     X = []
#     y = []
#     for file in files_list:
#         for i, id in enumerate(uniq_ID):
#             if id in file.split("/")[-1]:
#                 X.append(file)
#                 y.append(targets[i])
#
#     # Create the splits
#     for i in range(nbr_splits):
#         X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=percent_in_test, random_state=i)
#         splits_dict["split{}".format(i)] = [X_train, X_test, y_train, y_test]
#
# elif pairing_pn != "no" and pairing_12 == "no":  # pairing of positive and negative files
#     X_pos = []
#     y_pos = []
#     for file in files_list:
#         if pair_id_pos in file.split("/")[-1]:
#             for i, id in enumerate(uniq_ID):
#                 if id in file.split("/")[-1]:
#                     X_pos.append(file)
#                     y_pos.append(targets[i])
#
#     # Create the splits
#     for i in range(nbr_splits):
#         X_train_p, X_test_p, y_train_p, y_test_p = train_test_split(X_pos, y_pos, test_size=percent_in_test,
#                                                                     random_state=i)
#         X_train_n = [i.replace(pair_id_pos, pair_id_neg) for i in X_train_p]
#         X_test_n = [i.replace(pair_id_pos, pair_id_neg) for i in X_test_p]
#         y_train_n = []
#         y_test_n = []
#         for file in X_train_n:
#             for i, id in enumerate(uniq_ID):
#                 if id in file.split("/")[-1]:
#                     y_train_n.append(targets[i])
#         for file in X_test_n:
#             for i, id in enumerate(uniq_ID):
#                 if id in file.split("/")[-1]:
#                     y_test_n.append(targets[i])
#
#         # shuffle datasets to make sure pos files and neg are seen in a random order by the algorithms
#         X_train = X_train_p + X_train_n
#         y_train = y_train_p + y_train_n
#         train_zip = list(zip(X_train, y_train))
#         random.Random(13).shuffle(train_zip)
#         X_train, y_train = zip(*train_zip)
#
#         X_test = X_test_p + X_test_n
#         y_test = y_test_p + y_test_n
#         test_zip = list(zip(X_test, y_test))
#         random.Random(13).shuffle(test_zip)
#         X_test, y_test = zip(*test_zip)
#
#         splits_dict["split{}".format(i)] = [X_train, X_test, y_train, y_test]
#
# elif pairing_pn == "no" and pairing_12 != "no":  # pairing over another condition
#     X_1 = []
#     y_1 = []
#     for file in files_list:
#         if pair_id_1 in file.split("/")[-1]:
#             for i, id in enumerate(uniq_ID):
#                 if id in file.split("/")[-1]:
#                     X_1.append(file)
#                     y_1.append(targets[i])
#
#     # Create the splits
#     for i in range(nbr_splits):
#         X_train_1, X_test_1, y_train_1, y_test_1 = train_test_split(X_1, y_1, test_size=percent_in_test,
#                                                                     random_state=i)
#         X_train_2 = [i.replace(pair_id_1, pair_id_2) for i in X_train_1]
#         X_test_2 = [i.replace(pair_id_1, pair_id_2) for i in X_test_1]
#         y_train_2 = []
#         y_test_2 = []
#         for file in X_train_2:
#             for i, id in enumerate(uniq_ID):
#                 if id in file.split("/")[-1]:
#                     y_train_2.append(targets[i])
#         for file in X_test_2:
#             for i, id in enumerate(uniq_ID):
#                 if id in file.split("/")[-1]:
#                     y_test_2.append(targets[i])
#
#         # shuffle datasets to make sure pos files and neg are seen in a random order by the algorithms
#         X_train = X_train_1 + X_train_2
#         y_train = y_train_1 + y_train_2
#         train_zip = list(zip(X_train, y_train))
#         random.Random(13).shuffle(train_zip)
#         X_train, y_train = zip(*train_zip)
#
#         X_test = X_test_1 + X_test_2
#         y_test = y_test_1 + y_test_2
#         test_zip = list(zip(X_test, y_test))
#         random.Random(13).shuffle(test_zip)
#         X_test, y_test = zip(*test_zip)
#
#         splits_dict["split{}".format(i)] = [X_train, X_test, y_train, y_test]
#
# elif pairing_pn != "no" and pairing_12 != "no":  # pairing pos/neg AND another condition
#     X_pos_1 = []
#     y_pos_1 = []
#     for file in files_list:
#         if pair_id_pos in file.split("/")[-1] and pair_id_1 in file.split("/")[-1]:
#             for i, id in enumerate(uniq_ID):
#                 if id in file.split("/")[-1]:
#                     X_pos_1.append(file)
#                     y_pos_1.append(targets[i])
#
#     # Create the splits
#     for i in range(nbr_splits):
#         X_train_pos_1, X_test_pos_1, y_train_pos_1, y_test_pos_1 = train_test_split(X_pos_1, y_pos_1,
#                                                                                     test_size=percent_in_test,
#                                                                                     random_state=i)
#         X_train_pos_2 = [i.replace(pair_id_1, pair_id_2) for i in X_train_pos_1]
#         X_test_pos_2 = [i.replace(pair_id_1, pair_id_2) for i in X_test_pos_1]
#         y_train_pos_2 = []
#         y_test_pos_2 = []
#         for file in X_train_pos_2:
#             for i, id in enumerate(uniq_ID):
#                 if id in file.split("/")[-1]:
#                     y_train_pos_2.append(targets[i])
#         for file in X_test_pos_2:
#             for i, id in enumerate(uniq_ID):
#                 if id in file.split("/")[-1]:
#                     y_test_pos_2.append(targets[i])
#
#         X_train_neg_1 = [i.replace(pair_id_pos, pair_id_neg) for i in X_train_pos_1]
#         X_test_neg_1 = [i.replace(pair_id_pos, pair_id_neg) for i in X_test_pos_1]
#         y_train_neg_1 = []
#         y_test_neg_1 = []
#         for file in X_train_neg_1:
#             for i, id in enumerate(uniq_ID):
#                 if id in file.split("/")[-1]:
#                     y_train_neg_1.append(targets[i])
#         for file in X_test_neg_1:
#             for i, id in enumerate(uniq_ID):
#                 if id in file.split("/")[-1]:
#                     y_test_neg_1.append(targets[i])
#
#         X_train_neg_2 = [i.replace(pair_id_1, pair_id_2) for i in X_train_neg_1]
#         X_test_neg_2 = [i.replace(pair_id_1, pair_id_2) for i in X_test_neg_1]
#         y_train_neg_2 = []
#         y_test_neg_2 = []
#         for file in X_train_neg_2:
#             for i, id in enumerate(uniq_ID):
#                 if id in file.split("/")[-1]:
#                     y_train_neg_2.append(targets[i])
#         for file in X_test_neg_2:
#             for i, id in enumerate(uniq_ID):
#                 if id in file.split("/")[-1]:
#                     y_test_neg_2.append(targets[i])
#
#         # shuffle datasets to make sure pos files and neg are seen in a random order by the algorithms
#         X_train = X_train_pos_1 + X_train_pos_2 + X_train_neg_1 + X_train_neg_2
#         y_train = y_train_pos_1 + y_train_pos_2 + y_train_neg_1 + y_train_neg_2
#         train_zip = list(zip(X_train, y_train))
#         random.Random(13).shuffle(train_zip)
#         X_train, y_train = zip(*train_zip)
#
#         X_test = X_test_pos_1 + X_test_pos_2 + X_test_neg_1 + X_test_neg_2
#         y_test = y_test_pos_1 + y_test_pos_2 + y_test_neg_1 + y_test_neg_2
#         test_zip = list(zip(X_test, y_test))
#         random.Random(13).shuffle(test_zip)
#         X_test, y_test = zip(*test_zip)
#
#         splits_dict["split{}".format(i)] = [X_train, X_test, y_train, y_test]



import numpy as np
import pandas as pd
import os
import base64
import io
from .Utils import *


class DataFormat:
    """
    Take data file(s) as input and output a matrix where columns are samples and lines features. With the matrix comes
    a list of the columns names to retrieve the samples properly.
    """

    def __init__(self, filename, data=None, use_raw=False, from_base64_str=True):
        self.use_raw = use_raw
        self.filename = filename
        self.base64 = from_base64_str
        self.data = data

        # TODO : make sure to check if "not progen" matrix are well handled
        if self.base64:
            self.in_format = "base64"
        elif os.path.isfile(filename):
            self.in_format = "file"
        elif os.path.isdir(filename):
            self.in_format = "LDTD"
        else:
            raise TypeError(
                "The given path is not valid, it has to be a file or a directory."
            )

    def convert(self):
        """
        Convert data format for easier use by the PICO
        """
        if self.in_format == "base64":
            data_type, data_string = self.data.split(",")
            self.data = base64.b64decode(data_string)
            data = self._convert_from_file()
        elif self.in_format == "file":
            data = self._convert_from_file()
        elif self.in_format == "LDTD":
            raise ValueError("reading data from LDTD is not supported yet")
            #data = self._convert_from_LDTD()
        else:
            raise ValueError("self.in_format does not correspond to accepted values (base64, file)")
        return data

    def _convert_from_file(self):
        """
        Take a file path or an StringIO object and read it as a pandas Dataframe
        """
        file_ext = self.filename.split(".")[-1]
        # TODO : beware of the sep (, or ;)
        if "csv" in file_ext:  # Abundance matrices of Progenesis are always in csv format, so its checked first
            if self.in_format == "base64":  # this condition is to make readable the input data from dcc.Upload
                self.data = io.StringIO(self.data.decode("utf-8"))
            else:  # this else is to enable the pd dataframe to be read from full file path
                self.data = self.filename
            header = pd.read_csv(self.data, header=None, sep=None, engine="python", nrows=3,
                                  index_col=0,).fillna("").to_numpy()

            # Needs to reset the pointer to the top of the ioString (to be able to read the string again)
            if self.in_format == "base64":
                self.data.seek(0)

            if "Normalised abundance" in header[0] or "Raw abundance" in header[0]:
                datatable = pd.read_csv(self.data, header=[0, 1, 2], sep=None, engine="python", index_col=0)
                # Will return : datatable_compoundsInfo, datatable, labels, sample_names
                return self._read_Progenesis_data_table(datatable, header)
            else:
                datatable = pd.read_csv(self.data, sep=None, engine="python", index_col=0)
                # WARNING : return None, datatable, None, None
                return self._read_general_data_table(datatable)

        elif ("xls" in file_ext or "od" in file_ext):  # TODO : restrict the "od" condition, might be too large
            if self.in_format == "base64":  # same as above
                self.data = io.StringIO(io.BytesIO(self.data))
            else:
                self.data = self.filename
            datatable = pd.read_excel(self.data, index_col=0)
            # WARNING : return None, datatable, None, None
            return self._read_general_data_table(datatable)

        else:
            raise TypeError("The input file is not of the right type, must be excel, odt or csv.")

    def _convert_from_LDTD(self):
        # TODO :  implement the handling of LDTD data format
        return ""

    def _read_general_data_table(self, datatable):
        """
        for now does nothing, but might be the place to deal with custom format of matrices with extra/unecessary columns
        or informations
        ! careful : output only the datable and 3 empty strings because the function that calls it only needs datatable,
        but that might change
        """
        # This return list fits the return when reading a Progenesis file
        return None, datatable, None, None

    def _read_Progenesis_data_table(self, datatable, header):
        """
        Assumes Raw data columns are written after Normalized data columns in the file.
        :param datatable:
        :return:
        """
        # print(header)
        if not self.use_raw and "Normalised abundance" in header[0]:  # header.columns.tolist():
            start_data = list(header[0]).index("Normalised abundance")
        elif self.use_raw and "Raw abundance" in header[0]:  # header.columns.tolist():
            start_data = list(header[0]).index("Raw abundance")
        else:
            raise KeyError("There is no Raw or Normalized abundance detected in the header.")

        new_header = []
        for l in header:
            new_header.append(list_filler(l))

        datatable.columns = new_header
        datatable_compoundsInfo = datatable.iloc[:, 0:start_data]
        datatable_compoundsInfo.columns = datatable_compoundsInfo.columns.droplevel([0, 1])
        datatable_compoundsInfo = datatable_compoundsInfo.T

        if self.use_raw:
            datatable = datatable["Raw abundance"]
            labels, sample_names = list(zip(*datatable.columns))
        else:
            datatable = datatable["Normalised abundance"]
            labels, sample_names = list(zip(*datatable.columns))

        datatable.columns = datatable.columns.droplevel(0)
        datatable = datatable.T

        datatable = datatable.loc[[index for index in datatable.index if "QC" not in index]]

        return datatable_compoundsInfo, datatable, labels, sample_names

        # start_normalized = header.columns.tolist().index("Normalised abundance")
        # labels_array = np.array(header.iloc[0].tolist())

        # if with_raw:
        #     start_raw = header.columns.tolist().index("Raw abundance")
        #     sample_names = datatable.iloc[:, start_normalized:start_raw].columns
        #     labels = labels_array.tolist()[start_normalized:start_raw]
        # else:
        #     sample_names = datatable.iloc[:, start_normalized:].columns
        #     labels = labels_array.tolist()[start_normalized:]
        #
        # current_label = ""
        # for idx, l in enumerate(labels):
        #     if l != "nan":
        #         current_label = l
        #     else:
        #         labels[idx] = current_label
        #
        # if with_raw:
        #     datatable_compoundsInfo = datatable.iloc[:, 0:start_normalized]
        #     datatable_normalized = datatable.iloc[:, start_normalized:start_raw]
        #     datatable_raw = datatable.iloc[:, start_raw:]
        #     datatable_raw.columns = [i.rstrip(".1") for i in datatable_raw.columns]  # Fix the columns names
        #
        #     datatable_normalized = datatable_normalized.T
        #     datatable_raw = datatable_raw.T
        #     datatable_compoundsInfo = datatable_compoundsInfo.T
        #     datatable_normalized.rename(columns={"Compound": "Sample"})
        #     datatable_raw.rename(columns={"Compound": "Sample"})
        #
        #     if self.use_raw:
        #         return datatable_compoundsInfo, datatable_raw, labels, sample_names
        #     else:
        #         return datatable_compoundsInfo, datatable_normalized, labels, sample_names
        # else:
        #     datatable_compoundsInfo = datatable.iloc[:, 0:start_normalized]
        #     datatable_normalized = datatable.iloc[:, start_normalized:]
        #     datatable_normalized = datatable_normalized.T
        #     datatable_compoundsInfo = datatable_compoundsInfo.T
        #     datatable_normalized.rename(columns={"Compound": "Sample"})
        #     return datatable_compoundsInfo, datatable_normalized, labels, sample_names



import importlib
import os
from typing import List, Union

import sklearn

from .MetaboModel import MetaboModel
from ..conf.SupportedModels import LEARN_CONFIG
from ..service import Utils


# TODO: deals with methods names' that are used in Results (for example), how to retrieve features/importance/etc
class ModelFactory:
    def __init__(self):
        self._SEED = 42

    def create_supported_models(self, CV_type = "GS") -> dict:
        """
        CV_type: for now supports GS (gridsearch) and RS (randomsearch), the value will adapt the retrieval of the
        hyperparameter grid to explore.
        """
        supported_models = {}

        if CV_type == "GS":
            parameter_grid_value = "GridSearch"
        elif CV_type == "RS":
            parameter_grid_value = "RandomSearch"
        else:
            raise ValueError(f"Supported values are 'GS' and 'RS', given value is {CV_type}")

        for model_name, model_configuration in LEARN_CONFIG.items():
            supported_models[model_name] = MetaboModel(
                model_configuration["function"], model_configuration["ParamGrid"][parameter_grid_value],
                model_configuration["importance_attribute"]
            )
        return supported_models

    def create_custom_model(self, model_name: str, needed_imports: str, params_grid: dict,
                            importance_attribute: str = Utils.DEFAULT_IMPORTANCE_ATTRIBUTE) -> MetaboModel:
        """
        Create a custom model (not included by default) for the PICO to use
        """
        imports_list = needed_imports.split(".")
        model = Utils.get_model_from_import(imports_list, model_name)
        return MetaboModel(model, params_grid, importance_attribute)



import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from dash import html

from ..conf import parameters as cfg
from decimal import Decimal
import math


class Plots:
    def __init__(self, colors: str, render="svg"):
        self.colors = colors
        self.render_mode = render
        self.__default_stylesheet_for_cooc = [
            {
                "selector": "node",
                "style": {
                    "width": "mapData(size, 0, 1, 5, 50)",
                    "height": "mapData(size, 0, 1, 5, 50)",
                    "content": "data(label)",
                    'text-background-color': 'white',  # Set the background color for the text
                    'text-background-opacity': 1,  # Set the opacity of the text background
                    'text-background-padding': '2px',
                    "font-size": "6px",
                    "text-valign": "center",
                    "text-halign": "center",
                    'border-color': 'black',  # Set the border color
                    'border-width': '1px',  # Set the border width
                    'border-style': 'solid',  # Set the border style
                    'text-opacity': 0,
                    'background-color': '#00143E',
                }
            },
            {
                "selector": "edge",
                "style": {
                    "width": "mapData(size, 0, 1, 1, 5)",
                    'line-color': 'gray',
                }
            }
        ]


    def show_umap(self, umap_data, classes, algo, slider_value, sample_ids: list):
        val = [5, 10, 40, 100, "used", "all"]
        fig = px.scatter(
            umap_data,
            x=0,
            y=1,
            color=classes,
            color_continuous_scale=self.colors,
            title="",
            hover_name=sample_ids,
            render_mode=self.render_mode,
        )

        fig.update_layout(
            {
                "plot_bgcolor": "rgba(0, 0, 0, 0)",
                "paper_bgcolor": "rgba(0, 0, 0, 0)",
            },
            title="UMAP applied on top "
                  + str(val[slider_value])
                  + " features selected by "
                  + algo,
        )
        return fig

    def show_3D_umap(self, umap_data, classes, algo, slider_value, sample_ids: list):
        val = [5, 10, 40, 100, "used", "all"]
        fig = px.scatter_3d(
            umap_data,
            x=0,
            y=1,
            z=2,
            color=classes,
            color_continuous_scale=self.colors,
            title="",
            hover_name=sample_ids,
        )

        fig.update_layout(
            title="UMAP applied on top " + str(val[slider_value]) + " features selected by "+ algo,
            template='plotly_white',
                  )
        return fig

    def show_PCA(self, pca_data, pca_labels, classes, slider_value, algo, sample_ids: list):
        val = [5, 10, 40, 100, "used", "all"]
        fig = px.scatter(
            pca_data,
            labels=pca_labels,
            x=0,
            y=1,
            color=classes,
            color_continuous_scale=self.colors,
            title="",
            hover_name=sample_ids,
            render_mode=self.render_mode,
        )
        fig.update_layout(
            {
                "plot_bgcolor": "rgba(0, 0, 0, 0)",
                "paper_bgcolor": "rgba(0, 0, 0, 0)",
            },
            title="PCA applied on top "
                  + str(val[slider_value])
                  + " features selected by "
                  + algo,
        )
        return fig

    def show_3D_PCA(self, pca_data, pca_labels, classes, slider_value, algo, sample_ids: list):
        val = [5, 10, 40, 100, "used", "all"]
        fig = px.scatter_3d(
            pca_data,
            labels=pca_labels,
            x=0,
            y=1,
            z=2,
            color=classes,
            color_continuous_scale=self.colors,
            title="",
            hover_name=sample_ids,
        )
        fig.update_layout(
            template='plotly_white',
            title="PCA applied on top "
                  + str(val[slider_value])
                  + " features selected by "
                  + algo,
        )
        return fig

    def show_general_confusion_matrix(self, cm, labels, text, algo, split):
        fig = go.Figure(
            data=go.Heatmap(
                z=cm,
                x=labels,
                y=labels,
                colorscale=self.colors,
                showscale=False,
            )
        )
        fig = fig.update_traces(text=text, texttemplate="%{text}", hovertemplate=None)
        fig.update_layout(
            title="Confusion matrix of split " + str(split) + " by " + algo,
            xaxis_title="Prediciton",
            yaxis_title="Truth",
        )
        return fig

    def show_accuracy_all(self, df, algo):
        """
        plot the balanced accuracy for each split on train and test set
        df : generated from Results.produce_accuracy_plot_all()
        """
        if "splits" not in df.columns:
            raise RuntimeError(
                "To show the global balanced accuracies plot, the dataframe needs to have a 'splits' column"
            )
        if "balanced accuracies" not in df.columns:
            raise RuntimeError(
                "To show the global balanced accuracies plot, the dataframe needs to have a 'accuracies' column"
            )
        if "color" not in df.columns:
            raise RuntimeError(
                "To show the global balanced accuracies plot, the dataframe needs to have a 'color' column"
            )

        fig = px.bar(df, x="splits", y="balanced accuracies", color="color", barmode="group", 
                     )

        fig.update_yaxes(range=[0, 1.1])
        fig.update_layout(
            template='plotly_white'
        )
        return fig

    def show_exp_info_all(self, df: pd.DataFrame):
        """
        display in table the number of samples, per classes, in train/test, etc.
        df : generated from Results.produce_info_expe()
        """
        if "stats" not in df.columns:
            raise RuntimeError(
                "To show the global accuracies plot, the dataframe needs to have a 'stats' column"
            )
        if "numbers" not in df.columns:
            raise RuntimeError(
                "To show the global accuracies plot, the dataframe needs to have a 'numbers' column"
            )
        
        r = []
        for i in range(len(df.index)):
            r.append(html.Tr([html.Td(df.iloc[i, 0]), html.Td(df.iloc[i, 1])]))

        table_body = [html.Tbody(r)]
        return table_body

    def show_features_selection(self, df: pd.DataFrame, algo):
        """
        table of features used by all models (all split of an algorithm)
        ranked by most used first
        only display 10 most or used in at least 75%? of models ?
        df : generated from Results.produce_features_importance_table()
        """
        if "features" not in df.columns:
            raise RuntimeError(
                "To show the global accuracies plot, the dataframe needs to have a 'features' column"
            )
        if "times_used" not in df.columns:
            raise RuntimeError(
                "To show the global accuracies plot, the dataframe needs to have a 'times_used' column"
            )
        if "importance_usage" not in df.columns:
            raise RuntimeError(
                "To show the global accuracies plot, the dataframe needs to have a 'importance_usage' column"
            )

        fig = go.Figure(
            data=[
                go.Table(
                    header=dict(values=list(df.columns), align="center"),
                    cells=dict(
                        values=[
                            df.iloc[:10, :].features,
                            df.iloc[:10, :].times_used,
                            df.iloc[:10, :].importance_usage,
                        ],
                        align="center",
                    ),
                )
            ]
        )
        fig.update_layout(
            title="Table of top 10 features sorted by importance for " + algo
        )
        fig.update_layout(title="Table of top 10 features sorted by importance")
        return fig

    def show_split_metrics(self):
        """
        display in table the number of samples, per classes, in train/test, etc. for one split
        """
        return

    def show_metabolite_levels(self, features_data, algo, sample_name: list):
        """
        Plot in stripchart (boxplot with point and no box)
        (with a dropdown to select the metabolite, max of N? metabolite)
        And show the intensity of this metabolite/ this feature in each class (one box per class)
        """
        if features_data.shape[1] > cfg.max_used_features_to_show:
            fig = generate_empty_figure("Features used is to large to be shown.")
            return fig

        df_container = []
        for c in features_data.columns:
            if c != "targets":
                df_container.append(pd.DataFrame({
                    "features_name": [c] * len(features_data["targets"]),
                    "intensity": list(features_data[c]),
                    "targets": list(features_data["targets"]),
                    "sample_name": sample_name,
                }))
        
        df_aggregated = pd.concat(df_container, ignore_index=True)

        fig = px.strip(
            df_aggregated,
            x="features_name",
            y="intensity",
            color="targets",
            title="Abundance of {} in each sample by class for {}".format("all",
                                                                          # feature,
                                                                          algo
                                                                          ),
            hover_name="sample_name",
        )

        fig.update_layout(
            template='plotly_white'
        )
        return fig

    def show_heatmap_wrong_samples(self, data_train, data_test, samples_names, algos):
        fig = go.Figure()
        fig.add_trace(
            go.Heatmap(
                z=data_test, x=algos, y=samples_names, opacity=1, colorscale="Reds"
            )
        )
        fig.update_layout(
            title="Number of wrong prediction per sample in test sets for all splits"
        )

        return fig

    def show_heatmap_features_usage(self, df, importance_threshold):
        fig = go.Figure()
        fig.add_trace(
            go.Heatmap(
                z=df.values, x=df.columns, y=df.index, opacity=1, colorscale="blues"
            )
        )
        importance_threshold = round(importance_threshold, 3)
        fig.update_layout(title=f"Mean importance of features (>{importance_threshold}) for all splits")
        return fig

    def show_barplot_comparaison_algo(self, algos, train_acc, train_std, test_acc, test_std):
        fig = go.Figure(data=[
            go.Bar(name='Train', x=algos, y=train_acc, error_y=dict(type='data', array=train_std)),
            go.Bar(name='Test', x=algos, y=test_acc, error_y=dict(type='data', array=test_std))
        ])
        # Change the bar mode
        fig.update_layout(barmode='group', template='plotly_white')
        return fig

    def show_2d(self, data, classes, sample_names):
        fig = px.scatter(
            data,
            x=data.columns[0],
            y=data.columns[1],
            color=classes,
            color_continuous_scale=self.colors,
            title="",
            hover_name=sample_names,
            render_mode=self.render_mode,
        )

        fig.update_layout(
            {
                "plot_bgcolor": "rgba(0, 0, 0, 0)",
                "paper_bgcolor": "rgba(0, 0, 0, 0)",
            }
        )
        
        
        return fig

    def show_3d(self, data, classes, sample_names):
        fig = px.scatter_3d(
            data,
            x=data.columns[0],
            y=data.columns[1],
            z=data.columns[2],
            color=classes,
            color_continuous_scale=self.colors,
            title="",
            hover_name=sample_names,
        )

        fig.update_layout(
            template='plotly_white'
        )
        
        return fig

    @staticmethod
    def get_train_test_split_graph(nbr_element, slider_value, percent_test):
        m_element = nbr_element
        k_test = int(percent_test * m_element)

        def choose(k, n):
            if n > k:
                return 0
            if n < 0:
                return 0
            if k < 0:
                return 0
            return math.comb(k,n)

        def f(i, j):
            return choose(m_element - i, j - i) * choose(i, k_test - (j - i)) / choose(m_element, k_test)

        M = np.zeros((m_element, m_element))
        for a in range(1, m_element + 1):
            for b in range(1, m_element + 1):
                M[a - 1, b - 1] = f(a, b)

        V = np.zeros(m_element)
        V[k_test] = 1
        nbr_limit = 100
        valeurs = np.zeros(nbr_limit)

        for c in range(nbr_limit):
            valeurs[c] = V[m_element - 1]
            V = np.matmul(V, M)

        def threshold_at_99(selected):
            vector = np.zeros(m_element)
            vector[k_test] = 1
            for i in range(selected - 1):
                vector = np.matmul(vector, M)
            result = 0
            for i in range(m_element):
                result += vector[m_element - i - 1]
                if result > 0.9999:
                    return int((m_element - i - 1) / m_element * 100)

        fig = px.scatter(y=valeurs, x=[i for i in range(1, len(valeurs) + 1)],
                    labels={'x': 'Number of splits',
                            'y': 'Probability a'},
                            render_mode="svg"
                         )
        fig.add_trace(go.Scatter(x=[slider_value], y=[valeurs[slider_value - 1]], mode='markers',
                                 marker_symbol='circle',
                                 marker_size=10))
        fig.add_annotation(x=slider_value, y=valeurs[slider_value - 1],
                           text=f"({slider_value} splits, a={valeurs[slider_value - 1]:.2f}, b={threshold_at_99(slider_value)}%)",
                           showarrow=True, arrowhead=1)
        fig.update_layout(showlegend=False, template="plotly_white")
        return fig

    def get_default_stylesheet_for_cooc_graph(self):
        return self.__default_stylesheet_for_cooc

    def format_style_for_selected_node(self, node):
        return {
            "selector": 'node[id = "{}"]'.format(node['id']),
            'style': {
                'content': 'data(label)',
                'text-opacity': 1,
                'background-color': '#13BD00'
            }
        }

    def create_coocurence_graph(self, counter, mean_importance, number_of_split):
        def formatNode(node):
            return {'data': {'id': node, 'label': node, 'size': mean_importance[node]}}

        def formatEdge(edge):
            return {'data': {'source': edge[0], 'target': edge[1], 'size': counter[edge] / number_of_split}}

        seen_feature = set()
        dash_parameters = []
        for pair, weight in counter.items():
            x, y = pair
            if x not in seen_feature:
                dash_parameters.append(formatNode(x))
                seen_feature.add(x)
            if y not in seen_feature:
                dash_parameters.append(formatNode(y))
                seen_feature.add(y)
            dash_parameters.append(formatEdge(pair))

        return dash_parameters



def generate_empty_figure(text):
    # Create an empty figure
    fig = go.Figure()

    # Add annotation
    fig.add_annotation(
        text=text,
        xref="paper", yref="paper",
        x=0.5, y=0.5, showarrow=False,
        font=dict(size=20)
    )

    # Update layout to remove axes and grid
    fig.update_layout(
        xaxis=dict(showgrid=False, zeroline=False, visible=False),
        yaxis=dict(showgrid=False, zeroline=False, visible=False)
    )

    return fig


import math

import dash_bootstrap_components as dbc
import pandas
from dash import html, Output, Input, dash, State, dcc, callback_context, MATCH, ALL
from dash.dcc import send_file
from os.path import basename

from .MetaTab import MetaTab
from ...domain import Controller
from ...service import Utils, Plots, init_logger, log_exceptions

EXP_NAME = []
SPLIT_NUMBER_VALUES = {i: str(i) for i in range(0, 101, 10)}
SPLIT_NUMBER_VALUES.update({1: "1"})

NUMBER_OF_STEP_IN_CLASS_BALANCE = 5

CONFIG = {
    "toImageButtonOptions": {
        "format": "svg",  # one of png, svg, jpeg, webp
        "height": None,
        "width": None,
        "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
    }
}


class SplitsTab(MetaTab):
    def __init__(self, app: dash.Dash, metabo_controller: Controller):
        super().__init__(app, metabo_controller)
        self._logger = init_logger()

    def getLayout(self) -> dbc.Tab:
        _introductionNotice = html.Div(
            className="fig_group_all_width",
            children=[
                dbc.Card(
                    children=[
                        dbc.CardBody(
                            [
                                html.Div(
                                    "In this tab, you will create a setting file with all info necessary "
                                    "to run a machine learning experiment. The recommended order to proceed is : Files, Classification design, "
                                    "Define Splits, Sample Pairing, Class Balancing, Generate file."
                                ),
                            ]
                        ),
                    ]
                ),
            ],
        )

        __dataFile = html.Div(
            [
                dbc.Label("Data file(s) *", className="form_labels"),
                html.Div(
                    [
                        dcc.Upload(
                            id="upload_datatable",
                            children=[
                                dbc.Button(
                                    "Upload File",
                                    id="upload_datatable_button",
                                    # className="custom_buttons",
                                    color="outline-primary",
                                )
                            ],
                        ),
                        dcc.Loading(
                            id="upload_datatable_loading",
                            children=[
                                html.Div(
                                    id="upload_datatable_output",
                                    style={"color": "green"},
                                )
                            ],
                            style={"width": "100%"},
                            type="dot",
                            color="#13BD00",
                        ),
                    ],
                    style={"display": "flex", "align-items": "center"},
                ),
                dbc.FormText(
                    "You can give a Progenesis abundance file, or a matrix with samples as lines and features as "
                    "columns.",
                ),
                html.Div(id="error_upload_datatable", style={"color": "red"}),
            ],
            className="form_field",
            id="datatable-section",
            style={"display": "none"},
        )

        __metaDataFile = html.Div(
            [
                dbc.Label(
                    className="form_labels",
                    children=[
                        html.Span("Metadata file  "),
                        html.Span(
                            "(optionnal if Progenesis matrix given)",
                            style={"font-size": "0.9em", "text-transform": "none"},
                        ),
                    ],
                ),
                html.Div(
                    [
                        dcc.Upload(
                            id="upload_metadata",
                            children=[
                                dbc.Button(
                                    "Upload File",
                                    id="upload_metadata_button",
                                    # className="custom_buttons",
                                    color="outline-primary",
                                )
                            ],
                        ),
                        dcc.Loading(
                            id="upload_metadata_loading",
                            children=[
                                html.Div(
                                    id="upload_metadata_output",
                                    style={"color": "green"},
                                )
                            ],
                            style={"width": "100%"},
                            type="dot",
                            color="#13BD00",
                        ),
                    ],
                    style={"display": "flex", "align-items": "center"},
                ),
                dbc.FormText(
                    "The metadata file should at least contain : one column with samples name corresponding to names in the"
                    "data file, and one column of target/class/condition.",
                ),
                html.Div(id="error_upload_metadata", style={"color": "red"}),
            ],
            className="form_field",
            id="metadata-section",
            style={"display": "none"},
        )

        __useRawData = html.Div(
            [
                dbc.Label("DATA NORMALIZATION FOR PROGENESIS", className="form_labels"),
                dbc.FormText(
                    "If there is normalized and raw data in your file, you can choose which one to use. ",
                ),
                dbc.RadioItems(
                    id="in_use_raw",
                    options=[
                        {"label": "Raw", "value": "raw"},
                        {"label": "Normalized", "value": "normalized"},
                        {"label": "Not Progenesis", "value": "nap"},
                    ],
                    # if self.metabo_controller.is_data_raw() is None
                    # else (
                    #    "raw"
                    #    if not self.metabo_controller.is_data_raw()
                    #    else "normalized"
                    # ),
                    labelCheckedStyle={"color": "#13BD00"},
                ),
                html.Div(id="error_data_normalization", style={"color": "red"}),
            ],
            className="form_field",
        )

        __removeRTlessThan1min = html.Div(
            [
                dbc.Label(
                    "Remove features with RT lower than 1 min (progenesis only)", className="form_labels"
                ),
                dbc.FormText(
                    "We highly recommend to keep this as true, choose false at your own risks (see in documentation).",
                ),
                dbc.RadioItems(
                    id="in_remove_rt",
                    options=[
                        {"label": "True", "value": True, "disabled": True},
                        {"label": "False", "value": False, "disabled": True},
                    ],
                    labelCheckedStyle={"color": "#13BD00"},
                ),
                html.Div(id="warning_select_false", style={"color": "red"}),
            ],
            className="form_field",
            id="remove-rt-section",
            style={"display": "none"},
        )

        _file = html.Div(
            className="title_and_form",
            children=[
                html.H4(id="CreateSplits_paths_title", children="Files"),
                dbc.Form(
                    children=[
                        dbc.Col(
                            children=[
                                __useRawData,
                                __removeRTlessThan1min,
                                __dataFile,
                                __metaDataFile,
                            ]
                        ),
                    ]
                ),
            ],
        )

        __typeGroupLink = dbc.Card(
            [
                html.Div(
                    [
                        dbc.Label("Name of the targets column"),
                        dbc.Checklist(
                            id="in_target_col_name",
                            value=[],
                            # options=Utils.format_list_for_checklist(
                            #    self.metabo_controller.get_metadata_columns()
                            # ),
                            # value=[], #if self.metabo_controller.get_target_column() is None else [self.metabo_controller.get_target_column()],
                            inline=True,
                        ),
                    ],
                    className="form_field",
                ),
                html.Div(
                    [
                        dbc.Label("Name of the unique id column"),
                        dbc.RadioItems(
                            id="in_ID_col_name",
                            options=Utils.format_list_for_checklist(
                                self.metabo_controller.get_metadata_columns()
                            ),
                            value=self.metabo_controller.get_id_column(),
                            inline=True,
                        ),
                    ],
                    className="form_field",
                ),
                html.Div(
                    id="info_progenesis_loaded",
                    style={
                        "color": "grey",
                        "padding-left": "2em",
                        "font-style": "italic",
                    },
                ),
            ],
            body=True,
        )

        __labelDefinition = dbc.Card(
            id="",
            children=[
                html.Div(
                    [
                        dbc.Label("Type of classification"),
                        dbc.RadioItems(
                            id="in_classification_type",
                            value=0,
                            inline=True,
                            options=[
                                {"label": "Binary", "value": 0},
                                {"label": "Multiclass", "value": 1, "disabled": True},
                            ],
                        ),
                    ],
                    className="form_field",
                ),
                html.Div(
                    [
                        dbc.Label("Labels"),
                        dbc.FormText("It is good practice to define the control group as target 0 and the 'condition' group as target 1. "),
                        dbc.FormText("To do so, add a 0 or a 1 at the beginning of corresponding labels (ex: 0GroupA or 0_GroupA). "),
                        dbc.FormText("See more details in the documentation."),
                        html.Div(
                            className="fig_group_mini",
                            id="define_classes_desgn_exp",
                            children=[
                                dbc.Input(
                                    id="class1_name",
                                    type="text",
                                ),
                                dbc.Checklist(
                                    id="possible_groups_for_class1",
                                ),
                                dbc.Input(
                                    id="class2_name",
                                ),
                                dbc.Checklist(
                                    id="possible_groups_for_class2",
                                ),
                            ],
                        ),
                    ],
                    className="form_field",
                ),
                html.Div(id="error_classification_type", style={"color": "red"}),
                dbc.Button(
                    "Add",
                    id="btn_add_design_exp",
                    color="primary",
                    className="custom_buttons",
                    n_clicks=0,
                ),
                html.Div(id="output_btn_add_desgn_exp"),
            ],
            body=True,
        )

        _classificationDesigns = html.Div(
            className="title_and_form",
            children=[
                html.H4(id="Exp_desg_title", children="Define Classification designs"),
                dbc.Form(
                    children=[
                        dbc.Col(
                            children=[
                                dbc.FormText("Link each sample to its target/class."),
                                __typeGroupLink,
                                html.Div(id="error_link_each_sample_to_target", style={"color": "red"}),
                                html.Br(),
                                dbc.FormText("Classification Designs."),
                                dbc.Card(
                                    id="setted_classes_container",
                                    children=self._get_wrapped_classification_designs(),
                                    style={"display": "block", "padding": "1em"},
                                ),
                                html.Div(id="error_classification_designs", style={"color": "red"}),
                                dbc.FormText("Define labels and filter out samples."),
                                __labelDefinition,
                            ]
                        ),
                    ]
                ),
            ],
        )

        __samplePairing = html.Div(
            children=[
                dbc.FormText("Select the column you want to group."),
                dbc.Label("Column(s)"),
                html.Div(
                    id="pairing_columns",
                    children=[
                        dcc.Dropdown(
                            self.metabo_controller.get_metadata_columns(),
                            id="pairing_group_column",
                        )
                    ],
                ),
            ]
        )

        __classBalancing = html.Div(
            children=[
                html.H4(id="class_balancing_title", children="Class balancing"),
                html.Div(id="class_balancing_options", children=[]),
                html.Div(id="class_balancing_values_nutshell", children=[]),
            ]
        )

        _dataFusion = html.Div(
            className="title_and_form",
            children=[
                html.H4(id="sep_samples_title", children="Sample pairing"),
                dbc.Form(
                    children=[
                        dbc.Col(
                            [
                                __samplePairing,
                                __classBalancing
                            ]
                        )
                    ]
                ),
            ],
        )

        __sampleProportion = html.Div(
            [
                dbc.Label("Proportion of samples in test"),
                dbc.Input(
                    id="in_percent_samples_in_test",
                    value=self.metabo_controller.get_train_test_proportion(),
                    type="number",
                    min=0.01,
                    max=1,
                    step=0.01,
                    size="5",
                ),
                html.Div(id="error_percent_samples_in_test", style={"color": "red", "margin-top": "0.5em"}),
            ],
            className="form_field",
        )

        __trainTestSplitGraph = html.Div(
            [
                dbc.Label("Train/Test split graph"),
                dcc.Loading(
                    dcc.Graph(id="train_test_split_graph", config=CONFIG),
                    type="dot",
                    color="#13BD00",
                ),                
                dcc.Slider(
                    id='in_nbr_splits',
                    min=1,
                    max=100,
                    step=1,
                    value=self.metabo_controller.get_number_of_splits(),
                    marks=SPLIT_NUMBER_VALUES,
                ),
                html.Div(
                    children=[
                        html.P('a) Probability that all the samples are seen at least once in a test set'),
                        html.P(
                            'b) With a confidence level of 99.99%, proportion of the samples that are seen at least '
                            'once in a test set'),
                    ])
            ],
            className="form_field",
        )

        _splitDefinition = html.Div(
            className="title_and_form",
            children=[
                html.H4(id="Define_split_title", children="Define splits"),
                dbc.Form(
                    children=[
                        dbc.Col(children=[__sampleProportion, __trainTestSplitGraph]),
                    ]
                ),
            ],
        )
        __LDTDDataType = html.Div(
            [
                dbc.Label("Processing according to data type"),
                dbc.FormText(
                    "LDTD1 means the preprocessing will be done on all samples in one time. "
                    "LDTD2 means the preprocessing will be done seperatly for each split."
                ),
                dbc.RadioItems(
                    id="in_type_of_data",
                    value="none",
                    inline=True,
                    options=[
                        {"label": "None", "value": "none"},
                        {"label": "LDTD 1", "value": "LDTD1"},
                        {"label": "LDTD 2", "value": "LDTD2"},
                    ],
                ),
            ],
            className="form_field",
        )

        __LDTDPeakPicking = html.Div(
            [
                dbc.Label("Perform peak picking"),
                dbc.RadioItems(
                    id="in_peak_picking",
                    value=0,
                    inline=True,
                    options=[
                        {"label": "No", "value": 0, "disabled": True},
                        {"label": "Yes", "value": 1, "disabled": True},
                    ],
                ),
            ],
            className="form_field",
        )

        __LDTDAlignment = html.Div(
            [
                dbc.Label("Perform alignment"),
                dbc.RadioItems(
                    id="in_alignment",
                    value=0,
                    inline=True,
                    options=[
                        {"label": "No", "value": 0, "disabled": True},
                        {"label": "Yes", "value": 1, "disabled": True},
                    ],
                ),
            ],
            className="form_field",
        )

        __LDTDNormalization = html.Div(
            [
                dbc.Label("Perform normalization"),
                dbc.RadioItems(
                    id="in_normalization",
                    value=0,
                    inline=True,
                    options=[
                        {"label": "No", "value": 0, "disabled": True},
                        {"label": "Yes", "value": 1, "disabled": True},
                    ],
                ),
            ],
            className="form_field",
        )

        __peakThreshold = html.Div(
            [
                dbc.Label("Peak Threshold"),
                dbc.Input(
                    id="in_peak_threshold_value",
                    value="500",
                    type="number",
                    min=1,
                    size="5",
                ),
            ],
            className="form_field",
        )

        __autoOptimizeNumber = html.Div(
            [
                dbc.Label("AutoOptimize number"),
                dbc.Input(
                    id="in_autoOptimize_value",
                    value="20",
                    type="number",
                    min=1,
                    size="5",
                ),
            ],
            className="form_field",
        )

        # TODO: Not displayed
        _otherProcessing = html.Div(
            className="title_and_form",
            children=[
                html.H4(id="preprocess_title", children="Other Preprocessing"),
                dbc.Form(
                    children=[
                        dbc.Col(
                            children=[
                                dbc.FormText(
                                    "Options in case of LDTD data that needs to be preprocess"
                                ),
                                dbc.Collapse(
                                    dbc.Card(
                                        dbc.CardBody(
                                            children=[
                                                __LDTDDataType,
                                                __LDTDPeakPicking,
                                                __LDTDAlignment,
                                                __LDTDNormalization,
                                                __peakThreshold,
                                                __autoOptimizeNumber,
                                            ]
                                        )
                                    ),
                                    id="collapse_preprocessing",
                                ),
                                dbc.Button(
                                    "Open",
                                    id="collapse_preprocessing_button",
                                    className="custom_buttons",
                                    color="primary",
                                    n_clicks=0,
                                ),
                            ]
                        )
                    ]
                ),
            ],
            style={"visibility": "hidden"},
        )

        _generateFile = html.Div(
            className="title_and_form",
            children=[
                html.H4(id="create_split_title", children="E) Generate file"),
                dbc.Form(
                    children=[
                        dbc.Col(
                            children=[
                                html.Div(id="output_button_split_file"),
                                html.Div(
                                    className="button_box",
                                    children=[
                                        html.Div(
                                            "Before clicking on the Create button, make shure all field with an * are correctly filled."
                                        ),
                                        dbc.Button(
                                            "Create",
                                            color="primary",
                                            id="split_dataset_button",
                                            className="custom_buttons",
                                            n_clicks=0,
                                        ),
                                        html.Div(
                                            id="output_button_split",
                                            children="",
                                            style={"display": "none"},
                                        ),
                                    ],
                                ),
                            ]
                        )
                    ]
                ),
            ],
        )

        return dbc.Tab(
            className="global_tab",
            label="Splits",
            children=[
                _introductionNotice,
                html.Div(
                    className="fig_group",
                    children=[
                        _file,
                        _classificationDesigns,
                    ],
                ),
                html.Div(
                    className="fig_group", children=[_splitDefinition, _dataFusion]
                ),
                html.Div(
                    className="fig_group", children=[_otherProcessing, _generateFile]
                ),
            ],
        )

    def _registerCallbacks(self) -> None:
        @self.app.callback(
            Output("train_test_split_graph", "figure"),
            [
                Input("in_nbr_splits", "value"),
                Input("in_percent_samples_in_test", "value"),
                Input("upload_datatable_output", "children"),
            ],
        )
        @log_exceptions(self._logger)
        def update_train_test_split_graph(slider_value, percent_test, uploaded_data):
            """
            uploaded_data : not used in the function, but its presence is needed to triger the callback
                            when the data file is uploaded/changed
            """
            if percent_test is None:
                return dash.no_update
            nbr_of_samples = len(self.metabo_controller.get_samples_id())
            if nbr_of_samples == 0:
                return dash.no_update
            percent_test = float(percent_test)
            slider_value = int(slider_value)

            return Plots.get_train_test_split_graph(
                nbr_of_samples, slider_value, percent_test
            )

        @self.app.callback(
            Output("in_use_raw", "value"), [Input("custom_big_tabs", "active_tab")]
        )
        @log_exceptions(self._logger)
        def update_use_raw(active_tab):
            if active_tab == "tab-1":
                use_raw = self.metabo_controller.is_data_raw()
                if use_raw is None:
                    return None
                if use_raw:
                    return "raw"
                progenesis_usage = self.metabo_controller.is_progenesis_data()
                if progenesis_usage:
                    return "normalized"
                return "nap"
            else:
                return dash.no_update

        @self.app.callback(
            [Output("datatable-section", "style"),
             Output("remove-rt-section", "style"),
             Output("in_remove_rt", "options"),
             Output("in_remove_rt", "value")],
            [Input("in_use_raw", "value"), Input("custom_big_tabs", "active_tab")],
        )
        @log_exceptions(self._logger)
        def normalization_selection(value, active_tab):
            options = [
                {"label": "True", "value": True},
                {"label": "False", "value": False},
            ]
            disabled_options = []
            for option in options:
                copied_option = option.copy()
                copied_option["disabled"] = True
                disabled_options.append(copied_option)

            if value is not None:
                if value == "nap":  # Not a Progenesis file
                    self.metabo_controller.set_raw_use_for_data(False)
                    self.metabo_controller.set_data_matrix_remove_rt(False)
                    return {"display": "block"}, {"display": "block"}, disabled_options, None

                self.metabo_controller.set_raw_use_for_data(
                    True if value == "raw" else False
                )
                return {"display": "block"}, {"display": "block"}, options, True
            return dash.no_update

        @self.app.callback(
            Output("metadata-section", "style"),
            [Input("upload_datatable_output", "style")],
        )
        def display_metadata_after_data_upload(style):
            if style == {"color": "green"}:
                return {"display": "block"}
            return {"display": "none"}

        @self.app.callback(
            [
                Output("info_progenesis_loaded", "children"),
                Output("upload_datatable_output", "children"),
                Output("upload_datatable_output", "style"),
            ],
            [Input("upload_datatable", "contents"),
             Input("upload_metadata", "contents")],
            [State("upload_datatable", "filename")],
        )
        @log_exceptions(self._logger)
        def upload_data(list_of_contents, _, list_of_names):
            if callback_context.triggered[0]["prop_id"] == "upload_metadata.contents":
                return "", dash.no_update, dash.no_update
            if list_of_contents is not None:
                try:
                    self.metabo_controller.set_data_matrix_from_path(
                        list_of_names, data=list_of_contents
                    )
                except TypeError as err:
                    return dash.no_update, [html.P(str(err))], {"color": "red"}
                except pandas.errors.ParserError as err:
                    return (
                        dash.no_update,
                        [html.P("Rows must have an equal number of columns")],
                        {"color": "red"},
                    )
                self.metabo_controller.reset_classification_designs()

                if self.metabo_controller.is_progenesis_data():
                    # trigger the update of possible targets
                    return (
                        "Info: Selection not needed, handled by Progenesis.",
                        [html.P(f'"{list_of_names}" has successfully been uploaded !')],
                        {"color": "green"},
                    )
                return (
                    "",
                    [html.P(f'"{list_of_names}" has successfully been uploaded !')],
                    {"color": "green"},
                )
            else:
                return dash.no_update, dash.no_update, dash.no_update

        @self.app.callback(
            [
                Output("in_target_col_name", "options"),
                Output("in_ID_col_name", "options"),
                Output("pairing_group_column", "options"),
                Output("upload_metadata_output", "children"),
                Output("upload_metadata_output", "style"),
            ],
            [
                Input("upload_metadata", "contents"),
                Input("custom_big_tabs", "active_tab")
            ],
            [State("upload_metadata", "filename")],
        )
        @log_exceptions(self._logger)
        def get_metadata_cols_names_to_choose_from(list_of_contents, active_tab, list_of_names):
            triggered_item = callback_context.triggered[0]["prop_id"].split(".")[0]
            self._logger.info(f"triggered_item: [{triggered_item}]")
            if active_tab == "tab-1":
                if triggered_item == "upload_metadata":
                    try:
                        self.metabo_controller.set_metadata(list_of_names, data=list_of_contents)
                    except TypeError as err:
                        return [], [], [], html.P(str(err)), {"color": "red"}
                    except Exception as e:
                        return [], [], [], html.P(str(e)), {"color": "red"}
                    self.metabo_controller.reset_classification_designs()

                    formatted_columns = Utils.format_list_for_checklist(self.metabo_controller.get_metadata_columns())
                    return (formatted_columns, formatted_columns, formatted_columns,
                            html.P(f'"{list_of_names}" has successfully been uploaded !'), {"color": "green"})
                else:
                    formatted_columns = Utils.format_list_for_checklist(self.metabo_controller.get_metadata_columns())
                    return formatted_columns, formatted_columns, formatted_columns, dash.no_update, dash.no_update

            return dash.no_update, dash.no_update, dash.no_update, dash.no_update, dash.no_update

        @self.app.callback(
            Output("warning_select_false", "children"),
            [Input("in_remove_rt", "value")],
        )
        @log_exceptions(self._logger)
        def remove_features_from_datamatrix(value):
            if value is not None:
                self.metabo_controller.set_data_matrix_remove_rt(bool(value))
                if value:
                    return ""
                else:
                    return (
                        "Warning : features detected before 1 minute in the experiment are extremely likely to be "
                        "noise and to be biologically irrelevant."
                    )

        @self.app.callback(
            Output("define_classes_desgn_exp", "children"),
            [Input("in_classification_type", "value")],
        )
        @log_exceptions(self._logger)
        def define_classes_for_experiment_design(t):
            """
            if the classification type is binary (0), certain options will be available
            if it is multiclass (1), other options wil be shown
            :param t:
            :return:
            """
            if t == 0:
                return [
                    html.Div(
                        className="title_and_form_mini",
                        children=[
                            dbc.Form(
                                children=[
                                    dbc.Col(
                                        children=[
                                            html.Div(
                                                [
                                                    dbc.Label("Label 1"),
                                                    dbc.Input(
                                                        id="class1_name",
                                                        placeholder="Enter name",
                                                        debounce=True,
                                                        className="form_input_text",
                                                    ),
                                                ],
                                                className="form_field",
                                            ),
                                            html.Div(
                                                [
                                                    dbc.Label("Class(es)"),
                                                    dbc.Checklist(
                                                        id="possible_groups_for_class1"
                                                    ),
                                                ]
                                            ),
                                        ]
                                    )
                                ]
                            )
                        ],
                    ),
                    html.Div(
                        className="title_and_form_mini",
                        children=[
                            dbc.Form(
                                children=[
                                    dbc.Col(
                                        children=[
                                            html.Div(
                                                [
                                                    dbc.Label("Label 2"),
                                                    dbc.Input(
                                                        id="class2_name",
                                                        placeholder="Enter name",
                                                        debounce=True,
                                                        className="form_input_text",
                                                    ),
                                                ],
                                                className="form_field",
                                            ),
                                            html.Div(
                                                [
                                                    dbc.Label("Class(es)"),
                                                    dbc.Checklist(
                                                        id="possible_groups_for_class2"
                                                    ),
                                                ]
                                            ),
                                        ]
                                    )
                                ]
                            )
                        ],
                    ),
                ]

        @self.app.callback(
            [
                Output("possible_groups_for_class1", "options"),
                Output("possible_groups_for_class2", "options"),
                Output("output_btn_add_desgn_exp", "children"),
            ],
            [
                Input("in_target_col_name", "value"),
                Input("info_progenesis_loaded", "children"),
                Input("custom_big_tabs", "active_tab"),
            ],
        )
        @log_exceptions(self._logger)
        def update_possible_classes_exp_design(target_col, children, active_tab):
            triggered_id = callback_context.triggered[0]["prop_id"].split(".")[0]
            if active_tab == "tab-1":
                if triggered_id == "info_progenesis_loaded" or \
                        (triggered_id == "in_target_col_name" and target_col not in [None, "", []]):
                    # Give the MetaData attribute the values of either one or multiple column to create the targets
                    # for the experiment
                    if triggered_id == "in_target_col_name":
                        self.metabo_controller.set_target_columns(target_col)
                    # Format the targets list to get only the (maybe new) classes names to display
                    formatted_possible_targets = Utils.format_list_for_checklist(
                        self.metabo_controller.get_unique_targets()
                    )
                    return (
                        formatted_possible_targets,
                        formatted_possible_targets,
                        "",
                    )
                else:
                    return dash.no_update, dash.no_update, dash.no_update
            else:
                return dash.no_update, dash.no_update, dash.no_update

        @self.app.callback(
            [Output("class1_name", "value"),
             Output("possible_groups_for_class1", "value"),
             Output("class2_name", "value"),
             Output("possible_groups_for_class2", "value"),
             Output("setted_classes_container", "children"),
             Output("setted_classes_container", "style"),
             Output("error_classification_type", "children")],
            [Input("btn_add_design_exp", "n_clicks"),
             Input("remove_classification_design_button", "n_clicks"),
             Input("in_target_col_name", "value"),
             Input("info_progenesis_loaded", "children"),
             Input("custom_big_tabs", "active_tab")],
            [State("class1_name", "value"),
             State("possible_groups_for_class1", "value"),
             State("class2_name", "value"),
             State("possible_groups_for_class2", "value")],
        )
        @log_exceptions(self._logger)
        def add_n_reset_classes_exp_design(n_add, n_remove, target_col, children, active_tab, c1, g1, c2, g2):
            triggered_id = callback_context.triggered[0]["prop_id"].split(".")[0]

            if (triggered_id == "remove_classification_design_button" or triggered_id == "in_target_col_name"
                    or triggered_id == "info_progenesis_loaded"):
                self.metabo_controller.reset_classification_designs()
            elif triggered_id == "btn_add_design_exp":
                try:
                    self.metabo_controller.add_classification_design({c1: g1, c2: g2})
                    self._logger.info(f"Classification designs : {self.metabo_controller._metabo_experiment.classification_designs}")
                except ValueError as ve:
                    return (dash.no_update, dash.no_update, dash.no_update, dash.no_update,
                            dash.no_update, dash.no_update, str(ve))

            return ("", 0, "", 0, self._get_wrapped_classification_designs(), {"display": "block", "padding": "1em"}, "")

        @self.app.callback(
            Output("collapse_preprocessing", "is_open"),
            [Input("collapse_preprocessing_button", "n_clicks")],
            [State("collapse_preprocessing", "is_open")],
        )
        def toggle_collapse_preprocessing(n, is_open):
            if n:
                return not is_open
            return is_open

        @self.app.callback(
            Output("in_ID_col_name", "value"),
            [Input("in_ID_col_name", "value"), Input("custom_big_tabs", "active_tab")],
        )
        @log_exceptions(self._logger)
        def update_ID_col_name(new_value, active_tab):
            if active_tab == "tab-1":
                if new_value not in [None, ""]:
                    self.metabo_controller.set_id_column(new_value)
                return self.metabo_controller.get_id_column()
            return dash.no_update

        @self.app.callback(
            Output("pairing_group_column", "value"),
            [
                Input("pairing_group_column", "value"),
                Input("custom_big_tabs", "active_tab"),
            ],
        )
        @log_exceptions(self._logger)
        def update_pairing_group_column(new_value, active_tab):
            if active_tab == "tab-1":
                if new_value not in [None, ""]:
                    self.metabo_controller.set_pairing_group_column(new_value)
                return self.metabo_controller.get_pairing_group_column()
            return dash.no_update

        @self.app.callback(
            [
                Output("output_button_split_file", "children"),
                Output("error_percent_samples_in_test", "children"),
                Output("error_upload_datatable", "children"),
                Output("error_data_normalization", "children"),
                Output("error_upload_metadata", "children"),
                Output("error_classification_designs", "children"),
            ],
            [Input("split_dataset_button", "n_clicks")],
            [State("in_percent_samples_in_test", "value"),
             State("in_nbr_splits", "value")],
        )
        @log_exceptions(self._logger)
        def saving_params_of_splits_batch(n, train_test_proportion, nbr_splits):
            """
            Create the file (json) which will contains all info about the split creation / data experiment.
            """
            if n >= 1:
                train_test_proportion_error = ""
                try:
                    casted_train_test_proportion = float(train_test_proportion)
                    if casted_train_test_proportion >= 1 or casted_train_test_proportion <= 0:
                        train_test_proportion_error = "The proportion must be between 0 and 1 (excluded)."
                    self.metabo_controller.set_train_test_proportion(casted_train_test_proportion)
                except (ValueError, TypeError):
                    train_test_proportion_error = "The train/test proportion must be a decimal number between 0 and 1 " \
                                                  "excluded."

                normalization_error = ""
                datatable_error = ""
                metadata_error = ""
                invalid_id_error = ""
                classification_design_error = ""
                if self.metabo_controller.is_data_raw() is None:
                    normalization_error = "Please select a normalization method."
                elif not self.metabo_controller.data_is_set():
                    datatable_error = "You must upload a file before splitting it."
                elif not self.metabo_controller.metadata_is_set():
                    metadata_error = "You must upload a metadata file before splitting it."
                elif not self.metabo_controller.validate_id_column():
                    invalid_id_error = html.P(f'You must provide a valid name of unique id column', style={"color": "red"})
                elif not self.metabo_controller.get_all_classification_designs_names():
                    classification_design_error = "You must add at least one classification design before " \
                                                "splitting the data."

                if train_test_proportion_error != "" or datatable_error != "" or normalization_error != "" or metadata_error != "" \
                    or classification_design_error != "" or invalid_id_error != "":
                    return invalid_id_error, train_test_proportion_error, datatable_error, normalization_error, metadata_error, classification_design_error
                self.metabo_controller.set_number_of_splits(nbr_splits)
                self.metabo_controller.create_splits()
                
                # Dump file to dump folder and to save folder (backup)
                metabo_expe_filename = Utils.get_metabo_experiment_path("pico_splits") # Get save file path
                metabo_expe_obj = self.metabo_controller.generate_save()
                Utils.dump_metabo_expe(metabo_expe_obj) # Dump the classification design to the dump folder
                Utils.dump_metabo_expe(metabo_expe_obj, metabo_expe_filename) # Save the classification design
                del metabo_expe_obj

                self._logger.info(f"{self.metabo_controller._metabo_experiment.classification_designs}")
                self._logger.info(f"Classification design splits file '{metabo_expe_filename}' saved.")
                send_file(Utils.get_dumped_metabo_experiment_path())
                
                return (
                    [html.P(f'The parameters file "{basename(metabo_expe_filename)}" and the splits have been created.', style={"color": "green"}),
                     html.P('Please select the "Machine Learning" tab to continue.')],
                    "", "", "", "", "",
                )
            else:
                return (dash.no_update,) * 6

        @self.app.callback(
            Output("class_balancing_options", "children"),
            [Input("btn_add_design_exp", "n_clicks"),
             Input("remove_classification_design_button", "n_clicks")],
        )
        @log_exceptions(self._logger)
        def add_class_balancing_options(n_clicks_add, n_clicks_remove):
            """
            Add a class balancing option to the list of class balancing options.
            """
            global_classes_repartition = None
            if n_clicks_add >= 1:
                global_classes_repartition = self.metabo_controller.get_classes_repartition_for_all_experiment()

            if global_classes_repartition is None:
                return dash.no_update

            sliders = []
            for classes_name, classes_repartition in global_classes_repartition.items():
                if len(classes_repartition) > 2:
                    raise NotImplementedError("Only two classes are supported for now.")

                keys = list(classes_repartition.keys())
                total = sum(classes_repartition.values())
                class_a_repartition = math.floor(classes_repartition[keys[0]] / total * 100)
                class_b_repartition = math.ceil(classes_repartition[keys[1]] / total * 100)

                if class_a_repartition == class_b_repartition:
                    sliders.append(
                        html.Div(
                            [
                                html.H5(f"{keys[0]} versus {keys[1]}"),
                                html.P("The classes are already balanced."),
                            ]
                        )
                    )
                    continue

                if class_a_repartition < class_b_repartition:
                    class_a_repartition, class_b_repartition = class_b_repartition, class_a_repartition
                    keys[0], keys[1] = keys[1], keys[0]

                difference_to_balance = class_a_repartition - 50

                steps = Utils.get_closest_integer_steps(difference_to_balance)

                marks = {}
                for step in steps:
                    marks[step] = f"{class_a_repartition - step}:{class_b_repartition + step}"

                sliders.append(
                    html.Div(
                        [
                            html.H5(f"{keys[0]} versus {keys[1]}"),
                            html.P(f"The classes are not balanced, you can use the slider to balance them."),

                            dcc.Slider(
                                id={
                                    "type": "class_balancing_slider",
                                    "index": classes_name,
                                },
                                min=0,
                                max=difference_to_balance,
                                step=1,
                                value=0,
                                marks=marks
                            ),
                        ]
                    )
                )

            return sliders

        @self.app.callback(
            Output({"type": "class_balancing_slider", "index": MATCH}, "value"),
            Input({"type": "class_balancing_slider", "index": MATCH}, "value"),
            [State({"type": "class_balancing_slider", "index": MATCH}, "id")],
        )
        @log_exceptions(self._logger)
        def update_class_balancing_slider(value, id):
            """
            Update the class balancing slider.
            """
            if value is None:
                return dash.no_update

            self.metabo_controller.set_balance_correction_for_experiment(id["index"], value)
            return value

        @self.app.callback(
            Output("error_link_each_sample_to_target", "children"),
            [Input("in_ID_col_name", "value"),
             Input("custom_big_tabs", "active_tab"),
             Input("split_dataset_button", "n_clicks")],
        )
        @log_exceptions(self._logger)
        def validate_id_column(value, active_tab, n_clicks) -> dash._callback.NoUpdate | str:
            if active_tab != "tab-1":
                return dash.no_update

            if self.metabo_controller.validate_id_column():
                return ""
            
            id_column = self.metabo_controller.get_id_column()
            if id_column is None:
                error_msg = dash.no_update if n_clicks < 1 else "You must provide a valid name of unique id column"
            else:                            
                error_msg = f"'{id_column}' is not a valid name of unique id column"
            return error_msg

    def _get_wrapped_classification_designs(self):
        children_container = [html.Div("Classification design")]
        all_classification_designs = (
            self.metabo_controller.get_all_classification_designs_names()
        )

        if len(all_classification_designs) == 0:
            button = html.Div(
                dbc.Button(
                    "Reset",
                    className="custom_buttons",
                    id="remove_classification_design_button",
                ),
                style={"display": "none"},
            )
            return html.Div([html.P("No classification design setted yet."), button])

        for _, full_name in all_classification_designs:
            children_container.append(
                html.Div(
                    children=["- " + full_name],
                    style={
                        "display": "flex",
                        "justify-content": "space-between",
                        "align-items": "center",
                    },
                )
            )
        button = html.Div(
            dbc.Button(
                "Reset",
                className="custom_buttons",
                id="remove_classification_design_button",
            ),
            style={"textAlign": "right"},
        )
        children_container.append(button)
        return children_container



import matplotlib
matplotlib.use('agg')

from dash import html, dcc
from dash.dependencies import Input, Output
import dash
import dash_bootstrap_components as dbc
from flask import request, jsonify
import os
import signal

from .tabs import *
from ..domain import Controller

from pico.service import set_log_filename, init_logger
import threading

# Code for the logging
if threading.current_thread() is threading.main_thread():
    logger = set_log_filename()
    logger.info(f"Starting PICO")
else:
    logger = init_logger()
    logger.debug(f"New thread '{threading.current_thread().name}')")

# Launch dash app
app = dash.Dash(
    __name__,
    external_stylesheets=[dbc.themes.LUX],
    meta_tags=[{"name": "viewport", "content": "width=device-width"}],
)
app.title = "PICO"
server = app.server
# app.scripts.config.serve_locally = False
app.css.config.serve_locally = False
app.config.suppress_callback_exceptions = True
app.css.append_css({"external_url": "https://codepen.io/chriddyp/pen/bWLwgP.css"})

metabo_controller = Controller()
infoTab = InfoTab(app, metabo_controller)
splitsTab = SplitsTab(app, metabo_controller)
mLTab = MLTab(app, metabo_controller)
resultsTab = ResultsTab(app, metabo_controller)
resultsAggregatedTab = AggregatedResultsTab(app, metabo_controller)
interpretTab = InterpretTab(app, metabo_controller)

app.layout = html.Div(
    id="page",
    children=[
        html.Div(id="dataCache", children=[], style={"display": "none"}),
        html.Div(
            id="title_container",
            className="row",
            style={"display": "flex", "justify-content": "space-between", "align-items": "center"},
            children=[
                html.Div(
                    children=[
                        html.H1(id="title", children="PICO"),
                        html.Div(
                            children=[
                                html.P(
                                    "Pipeline for Interpretable", style={"color": "white", 'text-transform': 'uppercase', "margin-bottom": "0"}
                                ),
                                html.P(
                                    "Classification of Omics",
                                    style={"color": "white", 'text-transform': 'uppercase', "margin-bottom": "0"},
                                ),
                            ],
                            id="acronym",
                            style={"display": "flex", "justify-content": "center"},
                        ),
                    ],
                    id="title_bg",
                ),
                html.Div(  # Conteneur parent du bouton
                    children=[
                        html.Button("X", id="close-button", style={
                            "color": "white", "background-color": "darkred",
                            "border": "none", "border-radius": "50%", "width": "30px", "height": "30px",
                            "font-size": "20px", "line-height": "20px", "text-align": "center", "padding": "0",
                            "cursor": "pointer", "margin": "20px", "margin-top": "0px", "flex-shrink": "0"})
                    ],
                    style={"display": "flex", "justify-content": "flex-end", "align-items": "center", "flex": "1"}
                ),
                html.Div(id="message_close", style={"color": "white", "background-color": "darkred", "font-size": "24px", "display": "none"}),
            ],
        ),
        dcc.Location(id='url', refresh=True),
        html.Div(id='clientside-container', style={"display": "none"}),
        html.Div(
            id="main-content",
            children=[
                dbc.Tabs(
                    id="custom_big_tabs",
                    active_tab="tab-0",
                    className="global_tabs_container",
                    children=[
                        infoTab.getLayout(),
                        splitsTab.getLayout(),
                        mLTab.getLayout(),
                        #dbc.Tab(label="Splits", disabled=True),
                        #dbc.Tab(label="Machine Learning", disabled=True),
                        resultsTab.getLayout(),
                        resultsAggregatedTab.getLayout()
                        # interpretTab.getLayout()
                    ],
                )
            ],
        ),
    ],
)


app.clientside_callback(
    """
    function(n_clicks) {
        if (n_clicks > 0) {
            if (confirm("Do you want to close PICO application server?")) {
                window.close();
                fetch('/shutdown', {method: 'POST'})
                    .then(response => response.json())
                    .then(data => console.log(data));
                document.getElementById('message_close').innerHTML = 'Server stopped. Please close this web page.';
                document.getElementById('message_close').style.display = 'block';
            }
        }
    }
    """,
    Output('clientside-container', 'children'),
    Input('close-button', 'n_clicks')
)

@server.route('/shutdown', methods=['POST'])
def shutdown():
    try:
        logger.info("Shutting down PICO server...")
        shutdown_func = request.environ.get('werkzeug.server.shutdown')
        if shutdown_func:
            shutdown_func()
        else:
            os.kill(os.getpid(), signal.SIGINT)
        return jsonify({'message': 'PICO server is shutting down...'})
    except Exception as e:
        logger.error(f"Error during shutdown: {e}")
        return jsonify({'error': str(e)}), 500



DATA_MATRIX = "Data\\Matrix_normalised_pos.csv"

# Cas, Temoin, TC
EXPERIMENT_DESIGNS = {
    "Ctrl_vs_Case": {
        "classes": {"Controls": ["Temoin"], "Cases": ["Cas"]},
        "TestSize": 0.2,
    },
    # "Control vs TC":{
    #     "classes": {
    #         "Controles":["Cas"],
    #         "TC": ["TC"]
    #     },
    #     "TestSize": 0.2,
    # },
    # "TC vs Cas":{
    #     "classes": {
    #         "TC":["TC"],
    #         "Cases": ["Cas"]
    #     },
    #     "TestSize": 0.2,
    # },
    # "Control vs all":{
    #     "classes": {
    #         "Control":["Temoin"],
    #         "All": ["Cas", "TC"]
    #     },
    #     "TestSize": 0.2,
    # },
}



import os
import pickle
from typing import Generator, Tuple, Union

import pandas as pd

from . import MetaboExperiment
from .ExperimentDTO import ExperimentDTO
from .Results import *
from ..service import init_logger

ROOT_PATH = os.path.dirname(__file__)
DUMP_PATH = os.path.join(ROOT_PATH, os.path.join("dumps", "splits"))


class Controller:
    """
    Bottleneck class to interface between backend and frontend
    """
    def __init__(self, metabo_experiment: MetaboExperiment = None):
        self._logger = init_logger()
        if metabo_experiment is None:
            self._metabo_experiment = MetaboExperiment()
        else:
            self._metabo_experiment = metabo_experiment

    def set_metadata(self, filename: str, data=None, from_base64=True) -> None:
        """
        Set the metadata informations for the experiment
        filename: the filename of metadata
        data: matrix of information contained in the file
        from_base64: specify encoding
        """
        self._metabo_experiment.set_metadata_with_dataframe(filename=filename, data=data, from_base64=from_base64)

    def set_data_matrix_from_path(self, path_data_matrix, data=None, from_base64=True):
        """
        Set the data from file for the experiment
        path_data_matrix: filename or path of the file
        data: matrix of data
        from_base64: specify encoding
        """
        return self._metabo_experiment.set_data_matrix(path_data_matrix, data=data, from_base64=from_base64)

    def get_metadata_columns(self) -> list:
        """
        Retrieve the names of the columns in the metadata matrix
        """
        return self._metabo_experiment.get_metadata_columns()

    def get_unique_targets(self) -> list:
        """
        Retrieve a list of unique targets by applying a "set()" function to the complete list of targets.
        """
        return self._metabo_experiment.get_unique_targets()

    def add_classification_design(self, classes_design: dict):
        """
        add an classification design to the experience
        classes_design: which target.s against which for the prediction, and the names (classes) of the groups of target.s
        """
        self._metabo_experiment.add_classification_design(classes_design)

    def get_classification_designs(self):
        """
        Retrieve all classification designs for an experience.
        """
        return self._metabo_experiment.get_classification_designs()

    def all_classification_designs_names(self) -> Generator[Tuple[str, str], None, None]:
        """
        Retrieve all classification designs names for an experience.
        """
        return self._metabo_experiment.all_classification_designs_names()

    def get_all_classification_designs_names(self) -> List[Tuple[str, str]]:
        """
        Retrieve a list of classification designs names for an experience.
        """
        return list(self._metabo_experiment.all_classification_designs_names())

    def reset_classification_designs(self):
        """
        Delete all existing classification designs.
        """
        self._metabo_experiment.reset_classification_designs()

    def remove_classification_design(self, name: str):
        """
        !!! --- NOT USED --- !!!
        Remove an classification design
        """
        self._metabo_experiment.remove_classification_design(name)

    def get_samples_id_from_splits(self, nbr_split_list, design):
        """
        !!! --- NOT USED --- !!!
        Retrieve sample identification in each split from file
        nbr_split_list: nbr of splits
        design:
        """
        samples_list = []
        for s in nbr_split_list:
            with open(os.path.join(DUMP_PATH, design + "_split_{}.p".format(s)), "rb") as split_file:
                samples_list.append(pickle.load(split_file)[:2])  # append list of X_train & X_test samples names
        return samples_list

    def set_id_column(self, id_column: str):
        """
        Set the id_column attribute of the MetaData object
        It is the column in the metadata used to have a unique id for each line/item/sample
        id_column: string of the name of the column
        """
        self._metabo_experiment.set_id_column(id_column)

    def validate_id_column(self) -> bool:
        """
        Ensure all values in the data index column (unique_ids) are in the the metadata id_column
        """
        return self._metabo_experiment.validate_id_column()

    def set_selected_models(self, selected_models: list):
        """
        Set the self._selected_models attribute of MetaboExperiment with the list given in argument
        and for each Classification Design object initialize basics of Results instances
        selected_models: list of models to run during the experiment
        """
        self._metabo_experiment.set_selected_models(selected_models)

    def learn(self):  # TODO: Come back to put this function in the diagram, suspect its to big for now
        self._metabo_experiment.learn()

    def get_results(self, design_name: str, algo: str):
        """
        Retrieve the results of a specific algorithm for a specific classification design
        design_name: classification design's name
        algo: specific algo for which we want the results
        """
        return self._metabo_experiment.classification_designs[design_name].results[algo].results

    def get_all_results(self):
        """
        Retrieve, for each classification design that is done, the results dict corresponding
        """
        return self._metabo_experiment.get_all_updated_results()

    def add_custom_model(self, model_name: str, needed_imports: str, params_grid: dict, importance_attribute: str):
        """
        Add the information needed to run a custom model
        model_name: the name of the model
        needed_imports: the modules needed to run this model
        params_grid: the hyperparameters grid with which hyperparameters to test and for which values
        importance_attribute: the attribute of the model to use to measure the importance of features
        """
        self._metabo_experiment.add_custom_model(model_name, needed_imports, params_grid, importance_attribute)

    def get_all_algos_names(self) -> list:
        """
        Retrieve the list of names from default (supported) models and custom models
        """
        ret = self._metabo_experiment.get_all_algos_names()
        self._logger.info(f"get_all_algos_names: {ret}")
        return ret

    def set_cv_type(self, cv_type: str):
        """
        Set the type of Cross-Validation (cv) for the experiment
        """
        self._metabo_experiment.set_cv_type(cv_type)

    def get_cv_types(self) -> List[str]:
        """
        Retrieve the dict of possible cross-validation types supported by the PICO
        """
        return self._metabo_experiment.get_cv_types()

    def get_selected_cv_type(self) -> str:
        """
        Return the type of CV selected for this experiment
        """
        return self._metabo_experiment.get_selected_cv_type()

    def generate_save(self) -> ExperimentDTO:
        """
        Return an object ExperimentDTO
        (which is a holder of some MetaboExperiment attributes)
        """
        return self._metabo_experiment.generate_save()

    def is_save_safe(self, saved_metabo_experiment_dto: ExperimentDTO) -> bool:
        """
        Verify that the hash from the saved ExperimentDTO is the same from the current object
        """
        return self._metabo_experiment.is_save_safe(saved_metabo_experiment_dto)

    def full_restore(self, saved_metabo_experiment_dto: ExperimentDTO):
        """
        Restore an experiment from a saving (always? from file)
        """
        self._metabo_experiment.full_restore(saved_metabo_experiment_dto)

    def partial_restore(self, saved_metabo_experiment_dto: ExperimentDTO, filename_data: str, filename_metadata: str,
                        data=None, from_base64_data: bool = True, metadata=None, from_base64_metadata=True,):
        """
        Restore only the parameters of an experiment
        It allows the use of identical parameters for different sets of data and metadata
        """
        self._metabo_experiment.partial_restore(saved_metabo_experiment_dto, filename_data, filename_metadata, data,
                                                from_base64_data, metadata, from_base64_metadata,)

    def load_results(self, saved_metabo_experiment_dto: ExperimentDTO):
        """
        Init a new experiment (new metadata and data_matrix) and load saved results
        """
        self._metabo_experiment.load_results(saved_metabo_experiment_dto)

    def get_target_column(self) -> str:
        """
        Retrieve the _target_column attribute of metadata
        """
        return self._metabo_experiment.get_target_column()

    def get_id_column(self) -> str:
        """
        Retrieve the _id_column attribute of metadata
        """
        return self._metabo_experiment.get_id_column()

    def set_number_of_splits(self, number_of_splits: int):
        """
        Set the value of the MetaboExpe attribute _number_of_splits
        """
        self._metabo_experiment.set_number_of_splits(number_of_splits)

    def get_number_of_splits(self) -> int:
        """
        Retrieve the value of the attribute _number_of_splits
        """
        return self._metabo_experiment.get_number_of_splits()

    def set_train_test_proportion(self, train_test_proportion: float):
        """
        Set the value of the MetaboExpe attribute _train_test_proportion
        """
        self._metabo_experiment.set_train_test_proportion(train_test_proportion)

    def get_train_test_proportion(self) -> float:
        """
        Retrieve the value of the attribute _train_test_proportion
        """
        return self._metabo_experiment.get_train_test_proportion()

    def create_splits(self) -> None:
        """
        Check that Experiment parameters are set and then : create an instance of SplitGroup for each classification Design
        (The init of SplitGroup triggers the _compute_splits function)
        If test_split_seed is provided, then only this test split seed is computed.
        """
        self._metabo_experiment.create_splits()

    def create_test_split_from_seed(self, test_split_seed: int) -> None:
        """
        Check that Experiment parameters are set and then : create an instance of SplitGroup
        having one test split from provided seed for each classification Design
        (The init of SplitGroup triggers the _compute_splits function)
        """
        self._metabo_experiment.create_splits(test_split_seed)

    def get_selected_models(self) -> List[str]:
        """
        Retrieve the name of the selected models to use for the experiment
        """
        return self._metabo_experiment.get_selected_models()

    def is_progenesis_data(self) -> bool:
        """
        Return the bool indicating if the data given is of the progenesis format
        """
        return self._metabo_experiment.is_progenesis_data()

    def get_pairing_group_column(self) -> str:
        """
        Retrieve the name of the column to use for pairing the samples
        """
        return self._metabo_experiment.get_pairing_group_column()

    def set_pairing_group_column(self, pairing_group_column: str):
        """
        Set the value of the MetaboExpe attribute _pairing_group_column
        """
        self._metabo_experiment.set_pairing_group_column(pairing_group_column)

    def is_data_raw(self) -> Union[bool, None]:
        """
        return the bool indicating to use either raw or normalized data
        """
        return self._metabo_experiment.is_data_raw()

    def set_raw_use_for_data(self, use_raw_data: bool):
        """
        Set the bool value of whether to use the raw data from a progenesis matrix or the normalized data
        """
        self._metabo_experiment.set_raw_use_for_data(use_raw_data)

    def get_data_matrix_remove_rt(self) -> bool:
        """
        return the value of the _remove_rt attribute
        (if true, remove the features detected before 1 minute of Retention Time)
        """
        return self._metabo_experiment.get_data_matrix_remove_rt()

    def set_data_matrix_remove_rt(self, remove_rt: bool):
        """
        set the value of the _remove_rt attribute
        (if true, remove the features detected before 1 minute of Retention Time)
        """
        self._metabo_experiment.set_data_matrix_remove_rt(remove_rt)

    def get_cv_folds(self) -> int:
        """
        Return the attribute of MetaboExperiment of the number of Cross Validation folds
        Default is 5
        """
        return self._metabo_experiment.get_cv_folds()

    def set_cv_folds(self, cv_folds: int):
        """
        Set the number of Cross Validation folds, must be >=2
        """
        self._metabo_experiment.set_cv_folds(cv_folds)

    def update_classification_designs_with_selected_models(self):
        self._metabo_experiment.update_classification_designs_with_selected_models()

    def is_the_data_matrix_corresponding(self, data: str) -> bool:
        return self._metabo_experiment.is_the_data_matrix_corresponding(data)

    def is_the_metadata_corresponding(self, metadata: str) -> bool:
        return self._metabo_experiment.is_the_metadata_corresponding(metadata)

    def get_final_targets_values(self):
        return self._metabo_experiment.get_final_targets_values()

    def set_final_targets_values(self, targets_columns: List[str]):
        self._metabo_experiment.set_final_targets_values(targets_columns)

    def add_final_targets_col_to_dataframe(self):
        self._metabo_experiment.add_final_targets_col_to_dataframe()

    def data_is_set(self) -> bool:
        return self._metabo_experiment.data_is_set()

    def metadata_is_set(self) -> bool:
        return self._metabo_experiment.metadata_is_set()

    def set_target_columns(self, target_cols: List[str]) -> None:
        self._metabo_experiment.set_target_columns(target_cols)

    def set_multithreading(self, activate_multithreading: bool):
        self._metabo_experiment.set_multithreading(activate_multithreading)

    def get_cv_algorithm_configuration(self) -> list:
        return self._metabo_experiment.get_cv_algorithm_configuration()

    def set_cv_algorithm_configuration(self, cv_algorithm_configuration: list):
        self._metabo_experiment.set_cv_algorithm_configuration(cv_algorithm_configuration)

    def get_samples_id(self):
        """
        Samples name from data file
        """
        return self._metabo_experiment.get_samples_id()

    def get_classes_repartition_for_all_experiment(self) -> dict:
        return self._metabo_experiment.get_classes_repartition_for_all_experiment()

    def get_balance_correction_for_all_experiment(self) -> dict:
        return self._metabo_experiment.get_balance_correction_for_all_experiment()

    def set_balance_correction_for_experiment(self, classification_design_name: str, balance_correction: int) -> None:
        self._metabo_experiment.set_balance_correction_for_experiment(classification_design_name, balance_correction)

    def display_splits(self) -> None:
        """
        Display the classes repartition for each split of each classification design.

        This can be used for debugging purposes from the automate.py script for example.

        The output will be something like:
            classification design 'first_design' details:
                Data set repartition: 'B': 96 (84%) vs 'C': 18 (16%) (Balance corr=34%).
                Classes 'B' vs 'C' repartition in splits (All | Train | Test):
                Split #00: All=[34 (65%) vs 18 (35%)] | Train=[14 (52%) vs 13 (48%)] | Test=[20 (80%) vs 5 (20%)]
                Split #01: All=[35 (66%) vs 18 (34%)] | Train=[15 (52%) vs 14 (48%)] | Test=[20 (83%) vs 4 (17%)]
                ...
        """
        self._metabo_experiment.display_splits()